import asyncio
import threading
from time import sleep

# The `AsyncLoopThread` class is a subclass of `threading.Thread` that runs an asynchronous event loop until it
# is stopped.
class AsyncLoopThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._stop_event = threading.Event()


    def run(self):
        self._loop = self._target()
        asyncio.set_event_loop(self._loop)
        self._stop_task = self._loop.create_task(self.check_stop())
        if not self._loop.is_running():
            self._loop.run_forever()
        return


    async def check_stop(self):
        while not self.stopped:
            await asyncio.sleep(0)
        self.shutdown()

    def shutdown(self):
        tasks = asyncio.all_tasks(loop=self._loop)
        for task in tasks:
            task.cancel()
        self._loop.stop()

    def stop(self):
        self._stop_event.set()
        while self._loop.is_running():
            sleep(0.01)

    @property
    def stopped(self):
        return self._stop_event.is_set()


from abc import ABC, abstractmethod


def is_valid_port(port):
    if port is None:
        return
    try:
        port = int(port)
        assert 0 <= port <= 65535, "Port must be an integer or string between 0 and 65535."
    except:
        raise ValueError("Port must be an integer or string between 0 and 65535.")

def is_valid_ipv4(value):
    if isinstance(value, str):
        octets = value.split(".")
        assert len(octets) == 4, "IPv4 addresses must have 4 octets"
        for octet in octets:
            octet = int(octet)
            assert octet >= 0 and octet <= 255, f"Invalid octet: {octet}"
        return True
    return False


# The `Validator` class is an abstract base class that provides a descriptor for validating attribute
# values before setting them.

class Validator(ABC):

    def __set_name__(self, owner, name):
        self.private_name = f"_{name}"

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)


    @abstractmethod
    def validate(self, value):
        pass


# The `OneOf` class is a validator that checks if a given value is one of the specified options.
class OneOf(Validator):
    def __init__(self, *options, predicate=None):
        self.options = set(options)
        self.predicates = None
        self.predicate = None

        if isinstance(predicate, (list, tuple, set)):
            for fn in predicate:
                if not callable(fn):
                    raise ValueError("Predicates must be callable")
            self.predicates = predicate
        else:
            if predicate is not None and not callable(predicate):
                raise ValueError("Predicates must be callable")
            self.predicate = predicate

    def validate(self, value):
        if value not in self.options:
            raise ValueError(f'Expected {value!r} to be one of {self.options!r}')

        if self.predicates is not None:
            for fn in self.predicate:
                if not fn(value):
                    raise ValueError(
                        f'Expected {fn.__name__} to be true for {value!r}'
                    )
        elif self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate.__name__} to be true for {value!r}'
            )

class TypeOf(Validator):
    def __init__(self, *types, subtypes=False, predicate=None):
        self.subtypes = subtypes
        self.types = set(types)
        self.predicates = None
        self.predicate = None

        if isinstance(predicate, (list, tuple, set)):
            for fn in predicate:
                if not callable(fn):
                    raise ValueError("Predicates must be callable")
            self.predicates = predicate
        else:
            if predicate is not None and not callable(predicate):
                raise ValueError("Predicates must be callable")
            self.predicate = predicate

    def validate(self, value):
        if self.subtypes:
            is_subtype = False
            for t in self.types:
                if isinstance(value, t):
                    is_subtype = True
                    break
            if not is_subtype:
                raise ValueError(f'Expected {value!r} to be subtype of {self.types!r}')
        elif type(value) not in self.types:
            raise ValueError(f'Expected {value!r} to be type of {self.types!r}')

        if self.predicates is not None:
            for fn in self.predicate:
                if not fn(value):
                    raise ValueError(
                        f'Expected {fn.__name__} to be true for {value!r}'
                    )
        elif self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate.__name__} to be true for {value!r}'
            )


# The `Number` class is a validator that checks if a given value is a number within a specified range.
class Number(Validator):
    def __init__(self, minvalue=None, maxvalue=None, predicate=None):
        self.minvalue = minvalue
        self.maxvalue = maxvalue
        self.predicates = None
        self.predicate = None

        if isinstance(predicate, (list, tuple, set)):
            for fn in predicate:
                if not callable(fn):
                    raise ValueError("Predicates must be callable")
            self.predicates = predicate
        else:
            if predicate is not None and not callable(predicate):
                raise ValueError("Predicates must be callable")
            self.predicate = predicate


    def validate(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f'Expected {value!r} to be an int or float')
        if self.minvalue is not None and value < self.minvalue:
            raise ValueError(
                f'Expected {value!r} to be at least {self.minvalue!r}'
            )
        if self.maxvalue is not None and value > self.maxvalue:
            raise ValueError(
                f'Expected {value!r} to be no more than {self.maxvalue!r}'
            )
        if self.predicates is not None:
            for fn in self.predicate:
                if not fn(value):
                    raise ValueError(
                        f'Expected {fn.__name__} to be true for {value!r}'
                    )
        elif self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate.__name__} to be true for {value!r}'
            )

# The `String` class is a validator that checks if a given value is a string and satisfies certain
# size and predicate conditions.
class String(Validator):
    def __init__(self, minsize=None, maxsize=None, predicate=None):
        self.minsize = minsize
        self.maxsize = maxsize
        self.predicates = None
        self.predicate = None

        if isinstance(predicate, (list, tuple, set)):
            for fn in predicate:
                if not callable(fn):
                    raise ValueError("Predicates must be callable")
            self.predicates = predicate
        else:
            if predicate is not None and not callable(predicate):
                raise ValueError("Predicates must be callable")
            self.predicate = predicate

    def validate(self, value):
        if not isinstance(value, str):
            raise TypeError(f'Expected {value!r} to be an str')

        if self.minsize is not None and len(value) < self.minsize:
            raise ValueError(
                f'Expected {value!r} to be no smaller than {self.minsize!r}'
            )
        if self.maxsize is not None and len(value) > self.maxsize:
            raise ValueError(
                f'Expected {value!r} to be no bigger than {self.maxsize!r}'
            )
        if self.predicates is not None:
            for fn in self.predicate:
                if not fn(value):
                    raise ValueError(
                        f'Expected {fn.__name__} to be true for {value!r}'
                    )

        elif self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate.__name__} to be true for {value!r}'
            )



class Immutable(Validator):
    def __init__(self, predicate=None):
        self.predicates = None
        self.predicate = None
        if isinstance(predicate, (list, tuple, set)):
            for fn in predicate:
                if not callable(fn):
                    raise ValueError("Predicates must be callable")
            self.predicates = predicate
        else:
            if predicate is not None and not callable(predicate):
                raise ValueError("Predicates must be callable")
            self.predicate = predicate


    def __set__(self, obj, value):
        if hasattr(obj, self.private_name):
            raise AttributeError('Cannot reassign immutable attribute')
        self.validate(value)
        setattr(obj, self.private_name, value)


    def validate(self, value):
        if self.predicates is not None:
            for fn in self.predicate:
                if not fn(value):
                    raise ValueError(
                        f'Expected {fn.__name__} to be true for {value!r}'
                    )
        elif self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate.__name__} to be true for {value!r}'
            )


class Bool(Validator):
    def __init__(self, predicate=None):
        self.predicates = None
        self.predicate = None

        if isinstance(predicate, (list, tuple, set)):
            for fn in predicate:
                if not callable(fn):
                    raise ValueError("Predicates must be callable")
            self.predicates = predicate
        else:
            if predicate is not None and not callable(predicate):
                raise ValueError("Predicates must be callable")
            self.predicate = predicate

    def __set__(self, obj, value):
        value = self.validate(value)
        setattr(obj, self.private_name, value)


    def validate(self, value):
        if value in {False, "False", "false", "no", "No", 0}:
            return False
        elif value in {True, "True", "true", "yes", "Yes", 1}:
            return True
        else:
            raise ValueError(f'Expected {value!r} to be a boolean')


class ImmutableString(Immutable, String):
    def __init__(self):
        Immutable.__init__(self)
        String.__init__(self)

    def validate(self, value):
        Immutable.validate(self, value)
        String.validate(self, value)


# The `LoggedAccess` class provides logging functionality for accessing and updating attributes of an
# object.
class LoggedAccess:

    def __set_name__(self, owner, name):
        import logging
        logging.basicConfig(level=logging.INFO)
        self.public_name = name
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        value = getattr(obj, self.private_name)
        logging.info('Accessing %r giving %r', self.public_name, value) # type: ignore
        return value

    def __set__(self, obj, value):
        logging.info('Updating %r to %r', self.public_name, value) # type: ignore
        setattr(obj, self.private_name, value)


from .abstract import Cache

class InMemoryCache(Cache):
    __slots__ = ()

    def __init__(
        self,
        instance: object,
        **kwargs
    ):
        self._instance = instance
        super().__init__(backend="memory", **kwargs)

    def __contains__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)
        return self._instance._cache_memory_conn.__contains__(key)

    @Cache.deserialize
    def __getitem__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)

        value = self._instance._cache_memory_conn[key]
        if value is None:
            return value

        if self.options.compression:
            value = self._decompress(value)

        return value

    @Cache.serialize
    def __setitem__(self, key, value):
        if not key.endswith(":cache"):
            key = self._parse_key(key)

        if self.options.compression:
            value = self._compress(value)

        self._instance._cache_memory_conn[key] = value

    def __delitem__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)
        del self._instance._cache_memory_conn[key]

    def keys(self):
        return self._instance._cache_memory_conn.keys()

    def values(self):
        return tuple(self._instance._cache_memory_conn.values())

    def items(self):
        return tuple(self._instance._cache_memory_conn.items())

    def clear(self):
        return self._instance._cache_memory_conn.clear()

    def _cleanup(self):
        pass



import time
from abc import ABC, abstractmethod
from inspect import iscoroutinefunction
from functools import wraps
from atexit import register as _register
from asyncio import sleep

import sqlite3
from redislite import Redis

from yarl import URL

from ..objects import RatelimitOptions, RatelimitInMemoryCache


class Ratelimit(ABC):
    """
    Abstract base class for implementing rate limiting functionality.

    Args:
        backend (str | None): The backend to use for rate limiting. Defaults to None.
        key (str | None): The key to use for rate limiting. Defaults to None.
        conn (_Redis | None): The connection object for the backend. Defaults to None.
        per_host (bool): Whether to apply rate limiting per host. Defaults to False.
        per_endpoint (bool): Whether to apply rate limiting per endpoint. Defaults to True.
        cache_timeout (float | int): The cache timeout in seconds. Defaults to 300.
        check_frequency (float | int): The frequency at which to check the rate limit. Defaults to 15.
        sleep_duration (float | int): The duration to sleep between rate limit checks. Defaults to 0.01.
        **kwargs: Additional keyword arguments to configure the rate limit options.

    Attributes:
        options (RatelimitOptions): The rate limit options object.
    """

    __slots__ = ("options", "_instance", "_ratelimit_type", "_backend", "_cache", "_key", "_threadpool")

    def __init__(
        self,
        backend:         str | None        = None,
        key:             str | None        = None,
        conn:            Redis | None      = None,
        per_host:        bool              = False,
        per_endpoint:    bool              = True,
        cache_timeout:   float | int       = 300,
        check_frequency: float | int       = 15,
        sleep_duration:  float | int       = 0.01,
        raise_errors:    bool              = False,
        **kwargs,
    ):
        self._backend = backend
        self._key = key if isinstance(key, str) else "Session"

        kwargs = {
            "per_host":         per_host,
            "per_endpoint":     per_endpoint,
            "cache_timeout":    cache_timeout,
            "check_frequency":  check_frequency,
            "sleep_duration":   sleep_duration,
            "raise_errors":     raise_errors,
            **kwargs
        }
        if kwargs.get("ratelimit_options") is not None:
            kwargs = {**kwargs, **kwargs["ratelimit_options"]}
        elif kwargs.get("ratelimit") is not None:
            kwargs = {**kwargs, **kwargs["ratelimit"]}

        self.options = RatelimitOptions.from_backend(backend, **kwargs)

        if self._backend == "memory":
            self._instance.__class__._ratelimit_memory_conn = RatelimitInMemoryCache(options=self.options, default=self.default) # type: ignore

        elif self._backend == "redis":
            if not hasattr(self._instance, "_redis_conn") or not isinstance(self._instance._redis_conn, Redis):
                self._instance.__class__._redis_conn = conn or Redis(**self.options.redis_server_config())

        elif self._backend == "sqlite":
            assert self.options.db is not None, "Database file must be provided for SQLite cache."
            if not hasattr(self._instance, "_sqlite_conn") or not isinstance(self._instance._sqlite_conn, sqlite3.Connection):
                self._instance.__class__._sqlite_conn = conn or sqlite3.connect(self.options.db)
            if not hasattr(self, "_cursor"):
                self._cursor = self._instance._sqlite_conn.cursor()
            self._instance._sqlite_conn.row_factory = sqlite3.Row
            self._create_tables()

        _register(self._cleanup)

    def __repr__(self) -> str:
        keys = ", ".join(f"{key.lstrip('_')}={getattr(self, key)}" for key in self.__slots__)
        return f"<{self._backend.title()}{self.__class__.__name__} {keys}>"

    def __str__(self) -> str:
        keys = ", ".join(f"{key.lstrip('_')}={getattr(self, key)}" for key in self.__slots__)
        return f"<{self._backend.title()}{self.__class__.__name__} {keys}>"

    @abstractmethod
    def ok(self):
        pass

    @property
    def default(self):
        return None

    @property
    def cache(self):
        if self._backend == "memory":
            return self._instance._ratelimit_memory_conn
        elif self._backend == "redis":
            return self._instance._redis_conn # type: ignore
        elif self._backend == "sqlite":
            return self._instance._sqlite_conn # type: ignore

    def clear(self):
        if self._backend == "memory":
            return self.cache.clear()
        elif self._backend == "redis":
            return self.cache.flushdb()
        elif self._backend == "sqlite":
            return self.cache.clear()

    def _cleanup(self):
        if self._backend == "redis":
            self.cache._cleanup() # type: ignore
            self.cache.shutdown() # type: ignore

    @property
    def key(self):
        return self._key

    @property
    def current_timestampns(self):
        return time.time_ns()

    @property
    def now(self):
        return time.time()

    def _parse_url(self, url):
        try:
            url = URL(str(url))
        except:
            return None
        if self.options.per_host:
            url = f"{url.scheme}://{url.host}"
        elif self.options.per_endpoint:
            url = f"{url.scheme}://{url.host}{url.path}"
        return url

    def _parse_key(self, url=None, method=None, keys=None, **kwargs):
        url = self._parse_url(url)
        keys = keys if isinstance(keys, (list, tuple, set)) else []
        key = ":".join(str(value) for value in (self._key, method, *keys, url, "ratelimit") if value is not None)
        return key

    def _set_redis_key(self, key, func, *args, **kwargs):
        ret = func(key, *args, **kwargs)
        self.cache.expire(key, int(self.options.cache_timeout)) # type: ignore
        return ret

    def increment(self, url=None, method=None, keys=None, **kwargs):
        key = self._parse_key(url=url, method=method, keys=keys)
        while not self.ok(key): # type: ignore
            if self.options.raise_errors:
                raise InterruptedError("Rate limit exceeded.")
            time.sleep(self.options.sleep_duration)

    async def increment_async(self, url=None, method=None, keys=None, **kwargs):
        key = self._parse_key(url=url, method=method, keys=keys)
        while not self.ok(key): # type: ignore
            if self.options.raise_errors:
                raise InterruptedError("Rate limit exceeded.")
            await sleep(self.options.sleep_duration)


class RatelimitDecoratorMixin:
    __slots__ = ()

    def __call__(self, func):
        if iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(**kwargs): # type: ignore
                await self.increment_async(**kwargs) # type: ignore
                return await func(**kwargs)
        else:
            @wraps(func)
            def wrapper(**kwargs):
                self.increment(**kwargs) # type: ignore
                return func(**kwargs)
        return wrapper


try:
    from stem import Signal as _Signal
    from stem.control import Controller as _Controller
    _has_stem = True
except ImportError:
    _has_stem = False


if _has_stem:
    from os import getpid as _getpid
    from random import Random as _Random
    from itertools import cycle as _cycle
    from atexit import register as _register
    from time import sleep as _sleep
    from functools import wraps as _wraps
    from requests import Session as _Session

    from .useragents import UserAgents as _UserAgents
    from .vars import  IP_APIS as _IP_APIS


    def _check_tor_service():
        from psutil import process_iter as _process_iter
        if not any(process.name() == "tor" for process in _process_iter()):
            from psutil import (
                Popen as _Popen
            )
            from subprocess import (
                DEVNULL as _DEVNULL,
            )
            tor = _Popen([
                "/usr/bin/tor",
                "-f", "/etc/tor/torrc",
                "--runasdaemon", "1"
            ], stdout=_DEVNULL, stderr=_DEVNULL)
            _sleep(3)
            if not tor.is_running():
                return _check_tor_service()
        return

    _check_tor_service()


    def _new_id(func):
        @_wraps(func)
        def wrapper(self, *args, **kwargs):
            self._controller.authenticate(password=self.password)
            self._controller.signal(_Signal.NEWNYM)
            return func(self, *args, **kwargs)
        return wrapper



    class TorSession(_Session):
        RNG = _Random()

        """
        tor_ports = specify Tor socks ports tuple (default is (9150,), as the default in Tor Browser),
        if more than one port is set, the requests will be sent sequentially through the each port;
        tor_cport = specify Tor control port (default is 9151 for Tor Browser, for Tor use 9051);
        password = specify Tor control port password (default is None);
        autochange_id = number of requests via a one Tor socks port (default=5) to change TOR identity,
        specify autochange_id = 0 to turn off autochange Tor identity;
        threads = specify threads to download urls list (default=8).
        """

        def __init__(self, tor_ports=(9000, 9001, 9002, 9003, 9004), tor_cport=9051,
                    password=None, autochange_id=5, headers=None, random_user_agents=True, **kwargs):
            self.headers = headers if isinstance(headers, dict) else {}
            self._random_user_agents = random_user_agents
            self._tor_ports = tor_ports
            self._tor_cport = tor_cport
            self._password = password
            self.autochange_id = autochange_id
            self.ports = _cycle(tor_ports)
            super().__init__()

            self._start_controller()
            _register(lambda: self._controller.close())


        @property
        def tor_ports(self):
            return self._tor_ports


        @property
        def tor_cport(self):
            return self._tor_cport


        @property
        def password(self):
            return self._password


        def check_service(self):
            from psutil import process_iter
            if not any(process.name() == "tor" for process in process_iter()):
                from psutil import Popen
                from os import devnull
                tor = Popen([
                    "/usr/bin/tor",
                    "-f", "/etc/tor/torrc",
                    "--runasdaemon", "1"
                ], stdout=open(devnull, 'w'), stderr=open(devnull, 'w'))

                if not tor.is_running():
                    return self.check_service()
            return

        def new_id(func):
            def wrapper(self, *args, **kwargs):
                with _Controller.from_port(port=self.tor_cport) as controller:
                    controller.authenticate(password=self.password)
                    controller.signal(_Signal.NEWNYM)
                return func(self, *args, **kwargs)
            return wrapper


        def check_ip(self):
            my_ip = self.get(self.RNG.choice(_IP_APIS)).text
            return my_ip


        def _start_controller(self):
            self._controller = _Controller.from_port(port=self.tor_cport)


        @_new_id
        def request(self, method, url, headers=None, **kwargs):
            port = next(self.ports)

            # if using requests_tor as drop in replacement for requests remove any user set proxy
            if kwargs.__contains__("proxy"):
                del kwargs["proxy"]


            if kwargs.__contains__("proxies"):
                del kwargs["proxies"]


            proxy = {
                "http": f"socks5h://localhost:{port}",
                "https": f"socks5h://localhost:{port}",
            }
            if self._random_user_agents:
                headers = headers or {}
                headers["User-Agent"] = _UserAgents.user_agent
            try:
                resp = super().request(method, url, headers=headers, proxies=proxy, **kwargs)
            except Exception as e:
                print(e)
                return self.request(method, url, proxy=proxy, **kwargs)
            return resp


        def get(self, url, **kwargs):
            return self.request("GET", url, **kwargs)


        def post(self, url, **kwargs):
            return self.request("POST", url, **kwargs)


        def put(self, url, **kwargs):
            return self.request("PUT", url, **kwargs)


        def patch(self, url, **kwargs):
            return self.request("PATCH", url, **kwargs)


        def delete(self, url, **kwargs):
            return self.request("DELETE", url, **kwargs)


        def head(self, url, **kwargs):
            return self.request("HEAD", url, **kwargs)


__all__ = "RatelimitMixin", "ratelimit"

from inspect import signature

from .pyratelimit import PyRatelimitFactory, PyRatelimitDecoratorFactory
from .redis import RedisRatelimitFactory, RedisRatelimitDecoratorFactory
from .sqlite import SQLiteRatelimitFactory
from ..utils import get_valid_kwargs
from ..objects import Alias


def ratelimit_factory(instance, backend="memory", *args, **kwargs):
    keys = set(kwargs.keys())
    if any((alias := a) in keys for a in Alias.RATELIMIT_TYPE.value):
        type = kwargs.pop(alias)
    else:
        type = "slidingwindow"

    type = type.lower()
    type = Alias.validate_ratelimit_type(type)

    if backend in Alias.MEMORY.value:
        self = PyRatelimitFactory(instance, type=type, *args, **kwargs)
    elif backend in Alias.REDIS.value:
        self = RedisRatelimitFactory(instance, type=type, *args, **kwargs)
    elif backend in Alias.SQLITE.value:
        self =  SQLiteRatelimitFactory(instance, type=type, *args, **kwargs)
    else:
        raise NotImplementedError(f"Ratelimit backend {backend} is not implemented.")

    return self


class ratelimit:
    def __new__(cls, backend="memory", *args, **kwargs):
        keys = set(kwargs.keys())
        if any((alias := a) in keys for a in Alias.RATELIMIT_TYPE.value):
            type = kwargs.pop(alias)
        else:
            type = "slidingwindow"

        type = type.lower()
        type = Alias.validate_ratelimit_type(type)

        if backend in Alias.MEMORY.value:
            self = PyRatelimitDecoratorFactory(type=type, *args, **kwargs)
        elif backend in Alias.REDIS.value:
            self = RedisRatelimitDecoratorFactory(type=type, *args, **kwargs)
        else:
            raise NotImplementedError(f"Ratelimit backend {backend} is not implemented.")
        return self


class RatelimitMixin:
    def __init__(
        self,
        backend: str | None     = "memory",
        **kwargs
    ):
        self._backend = backend
        self._ratelimiter = ratelimit_factory(self, backend, **kwargs)
        kwargs["backend"] = backend

        if "kwargs" in signature(super().__init__).parameters:
            super().__init__(**kwargs)
        else:
            super().__init__(**get_valid_kwargs(super().__init__, kwargs))


    @property
    def ratelimiter(self):
        return self._ratelimiter

    @property
    def ratelimit(self):
        return self._ratelimiter

    @property
    def backend(self):
        return self._backend

    @property
    def ratelimit_type(self):
        return self._ratelimiter._ratelimit_type



import inspect
from dataclasses import dataclass
from time import perf_counter
from functools import wraps
from random import Random
from math import floor as _floor

from .vars import STATUS_CODES


_slidingwindow_target = lambda window, limit, n_tests, **kwargs: ((window / limit) * n_tests, window)
_tokenbucket_target = lambda capacity, fill_rate, n_tests, **kwargs: ((capacity / fill_rate * (n_tests - capacity)), capacity / fill_rate)
_leakybucket_target = lambda capacity, leak_rate, n_tests, **kwargs: ((n_tests - capacity) / leak_rate, leak_rate / capacity)
_fixedwindow_target = lambda window, limit, n_tests, **kwargs: ((n_tests / limit) * window, window)
_gcra_target = lambda period, limit, n_tests, **kwargs: ((n_tests - (capacity := _floor(limit / period))) * period + (limit if n_tests <= capacity else 0), period)
_target = lambda type, **kwargs: globals()[f"_{type}_target"](**kwargs)


@dataclass(slots=True, frozen=True)
class RatelimitResult:
    name: str
    execution_time: float
    target_time: float
    delta: float
    observed_delta: float
    passed: bool
    args: dict

def take(predicate, iterable):
    """
    Returns an iterator that yields elements from the iterable as long as the predicate is True.

    Args:
        predicate (function): A function that takes an element from the iterable and returns a boolean value.
        iterable (iterable): An iterable object.

    Yields:
        The elements from the iterable that satisfy the predicate.

    """
    for x in iterable:
        if predicate(x):
            yield x
        else:
            continue

def timer(func):
    """
    Decorator that measures the execution time of a function and compares it to a target time with a given delta.

    Args:
        func: The function to be timed.

    Returns:
        The wrapped function.

    """
    if inspect.iscoroutinefunction(func):
        @wraps(func)
        async def wrapper(n_tests, urls=None, min=0, max=5, type="slidingwindow", backend="memory", limit=5, window=1, capacity=5, fill_rate=5, leak_rate=5, period=5, **kwargs): # type: ignore
            start = perf_counter()
            args = await func(n_tests, urls=urls, min=min, max=max, type=type, backend=backend, limit=limit, window=window, capacity=capacity, fill_rate=fill_rate, leak_rate=leak_rate, period=period, **kwargs)
            end = perf_counter()

            target_time, delta = _target(type, limit=limit, window=window, capacity=capacity, fill_rate=fill_rate, leak_rate=leak_rate, period=period, n_tests=n_tests)
            execution_time = end - start
            observed_delta = execution_time - target_time
            test_passed = abs(observed_delta) <= delta
            print(f"{func.__name__} took {execution_time:.2f}s, Expected {target_time:.2f}s within {delta:.2f}s delta for {type}. Observed Delta ({observed_delta:.2f}s) {'Passed' if test_passed else 'Failed'}")
            return RatelimitResult(func.__name__, execution_time, target_time, delta, observed_delta, test_passed, args)
    else:
        @wraps(func)
        def wrapper(n_tests, urls=None, min=0, max=5, type="slidingwindow", backend="memory", limit=5, window=1, capacity=5, fill_rate=5, leak_rate=5, period=5, **kwargs):
            start = perf_counter()
            args = func(n_tests, urls=urls, min=min, max=max, type=type, backend=backend, limit=limit, window=window, capacity=capacity, fill_rate=fill_rate, leak_rate=leak_rate, period=period, **kwargs)
            end = perf_counter()

            target_time, delta = _target(type, limit=limit, window=window, capacity=capacity, fill_rate=fill_rate, leak_rate=leak_rate, period=period, n_tests=n_tests)
            execution_time = end - start
            observed_delta = execution_time - target_time
            test_passed = abs(observed_delta) <= delta
            print(f"{func.__name__} took {execution_time:.2f}s, Expected {target_time:.2f}s within {delta:.2f}s delta for {type}. Observed Delta ({observed_delta:.2f}s) {'Passed' if test_passed else 'Failed'}")
            return RatelimitResult(func.__name__, execution_time, target_time, delta, observed_delta, test_passed, args)
    return wrapper


def extract_args(type, kwargs):
    """
    Extracts and returns the arguments based on the given type.

    Args:
        type (str): The type of the session algorithm.
        kwargs (dict): The keyword arguments.

    Returns:
        dict: The extracted arguments based on the type.
    """
    if type == "slidingwindow":
        return {"window": kwargs.get("window"), "limit": kwargs.get("limit"), "n_tests": kwargs.get("n_tests")}
    elif type == "fixedwindow":
        return {"window": kwargs.get("window"), "limit": kwargs.get("limit"), "n_tests": kwargs.get("n_tests")}
    elif type == "tokenbucket":
        return {"capacity": kwargs.get("capacity"), "fill_rate": kwargs.get("fill_rate"), "n_tests": kwargs.get("n_tests")}
    elif type == "leakybucket":
        return {"capacity": kwargs.get("capacity"), "fill_rate": kwargs.get("leak_rate"), "n_tests": kwargs.get("n_tests")}
    elif type == "gcra":
        return {"period": kwargs.get("period"), "limit": kwargs.get("limit"), "n_tests": kwargs.get("n_tests")}
    return {}


def make_test(typename, dct=False):
    """
    Generate test parameters based on the given typename.

    Parameters:
        typename (str): The type of test to generate parameters for.
        dct (bool, optional): If True, return the parameters as a dictionary.
                              If False (default), return the parameters as individual values.

    Returns:
        tuple or dict: The generated test parameters. If dct is True, returns a dictionary,
                       otherwise returns a tuple.

    Raises:
        None

    """
    RNG = Random()
    if typename == "slidingwindow":
        window = round(RNG.uniform(1, 5), 1)
        limit = RNG.randint(int(window), int(window*3))
        n_tests = max(15, RNG.randint(int(limit*3), int(limit*5)))
        return (window, limit, n_tests) if not dct else dict(window=window, limit=limit, n_tests=n_tests)

    elif typename == "fixedwindow":
        window = round(RNG.uniform(1, 5), 1)
        limit = RNG.randint(int(window), int(window*3))
        n_tests = max(15, RNG.randint(int(limit*3), int(limit*5)))
        return (window, limit, n_tests) if not dct else dict(window=window, limit=limit, n_tests=n_tests)

    elif typename == "tokenbucket":
        capacity = RNG.randint(1, 6)
        fill_rate = RNG.randint(int(capacity*1.5), int(capacity*3))
        n_tests = max(15, RNG.randint(int(fill_rate*3), int(fill_rate*5)))
        return (capacity, fill_rate, n_tests) if not dct else dict(capacity=capacity, fill_rate=fill_rate, n_tests=n_tests)

    elif typename == "leakybucket":
        capacity = RNG.randint(1, 6)
        leak_rate = RNG.randint(int(capacity*1.5), int(capacity*3))
        n_tests = max(15, RNG.randint(int(leak_rate*3), int(leak_rate*5)))
        return (capacity, leak_rate, n_tests) if not dct else dict(capacity=capacity, leak_rate=leak_rate, n_tests=n_tests)

    elif typename == "gcra":
        period = round(RNG.uniform(0.001, 2), 1)
        limit = RNG.randint(1, 10)
        n_tests = max(15, RNG.randint(int(limit*1.5), int(limit*5)))
        return (period, limit, n_tests) if not dct else dict(period=period, limit=limit, n_tests=n_tests)


IMAGE_TYPES = ("jpeg", "png", "svg", "webp")


def get_valid_kwargs(func, kwargs):
    """Get the subset of non-None 'kwargs' that are valid params for 'func'"""
    sig_params = list(inspect.signature(func).parameters)
    return {k: v for k, v in kwargs.items() if k in sig_params and v is not None}


def generate_random_weights(xlen, weights):
    """
    Generate random weights for a given length.

    Args:
        xlen (int or sequence): The length of the desired weights or a sequence of values that length can be obtained from.
        weights (tuple or list): The initial weights.

    Returns:
        tuple: The generated random weights.

    Raises:
        TypeError: If xlen is not an int or a sequence of values.
        AssertionError: If the sum of given weights is not less than 1.
    """
    if not isinstance(xlen, int):
        if isinstance(xlen, (list, tuple, set, range)):
            xlen = len(xlen)
        else:
            raise TypeError("xlen must be an int representing the length of the desired weights or a sequence of values that length can be obtained from")

    if isinstance(weights[0], float):
        weights = tuple((i, x) for i,x in enumerate(weights))
    elif isinstance(weights[0], int):
        weights = tuple((i, x/100) for i,x in enumerate(weights))
    elif isinstance(weights[0][0], int):
        weights = tuple((x[0], x[1]/100) for x in weights)

    mapping = dict(weights)
    indexes = set(x[0] for x in weights)
    weights = tuple(x[1] for x in weights)
    weight_to_split = 1 - sum(weights)
    assert weight_to_split > 0, "sum of given weights must be less than 1"
    len_of_weights = xlen - len(weights)
    constant = weight_to_split / len_of_weights
    return tuple(constant if i not in indexes else mapping[i] for i in range(xlen))


class Urls:
    """
    A class that provides various URL generation methods.

    Attributes:
        RNG (Random): An instance of the Random class for generating random values.
        STATUS_CODES (tuple): A tuple of status codes.
        port (int): The port number for the URLs.
        BASE_URL (str): The base URL.
        IP_URL (str): The URL for retrieving IP information.
        UUID_URL (str): The URL for generating UUIDs.
        DELETE_URL (str): The URL for DELETE requests.
        GET_URL (str): The URL for GET requests.
        PATCH_URL (str): The URL for PATCH requests.
        POST_URL (str): The URL for POST requests.
        PUT_URL (str): The URL for PUT requests.
    """
    RNG = Random()
    STATUS_CODES = tuple(STATUS_CODES.keys())

    def __init__(self, base_url=None, port=80, secure=False):
        """
        Initializes a new instance of the Urls class.

        Args:
            base_url (str, optional): The base URL. Defaults to None.
            port (int, optional): The port number for the URLs. Defaults to 80.
            secure (bool, optional): Indicates whether the URLs should use HTTPS. Defaults to False.
        """
        secure = "s" if secure else ""
        self.port = port
        self.BASE_URL = base_url or f"http{secure}://localhost:{self.port}"
        self.IP_URL = f"{self.BASE_URL}/ip"
        self.UUID_URL = f"{self.BASE_URL}/uuid"
        self.DELETE_URL = f"{self.BASE_URL}/delete"
        self.GET_URL = f"{self.BASE_URL}/get"
        self.PATCH_URL = f"{self.BASE_URL}/patch"
        self.POST_URL = f"{self.BASE_URL}/post"
        self.PUT_URL = f"{self.BASE_URL}/put"


    def random(self):
        return self.RNG.choice((
            self.GET_URL,
            self.POST_URL,
            self.PUT_URL,
            self.PATCH_URL,
            self.DELETE_URL,
            self.UUID_URL,
            self.IP_URL
        ))

    @classmethod
    def google_search(cls, min=0, max=100, weights=None):
        return f"https://www.google.com/search?q={cls.RNG.randint(min, max)}"

    @classmethod
    def GOOGLE_URLS(cls, n_trials=1000, min=0, max=100):
        return (cls.google_search(min, max) for _ in range(n_trials))

    def RANDOM_URLS(self, n_trials=1000, min=0, max=50, weights=None):
        if weights is not None:
            return (self.GET_URL + f"?id={id}" for id in self.RNG.choices(range(min, max), k=n_trials, weights=generate_random_weights(max-min, weights)))
        return (self.GET_URL + f"?id={self.RNG.randint(min, max)}" for _ in range(n_trials))

    def GET_URLS(self, n_trials=1000, suffix=""):
        return (self.GET_URL + suffix for _ in range(n_trials))

    def POST_URLS(self, n_trials=1000):
        return (self.POST_URL for _ in range(n_trials))

    def PUT_URLS(self, n_trials=1000):
        return (self.PUT_URL for _ in range(n_trials))

    def PATCH_URLS(self, n_trials=1000):
        return (self.PATCH_URL for _ in range(n_trials))

    def DELETE_URLS(self, n_trials=1000):
        return (self.DELETE_URL for _ in range(n_trials))

    def UUID_URLS(self, n_trials=1000):
        return (self.UUID_URL for _ in range(n_trials))

    def IP_URLS(self, n_trials=1000):
        return (self.IP_URL for _ in range(n_trials))

    def DRIP_URLS(self, n_trials=1000, delay=0, duration=0.01, numbyte=10):
        return (
            f"{self.BASE_URL}/drip?&duration={duration_}&numbytes={numbytes_}&code={code_}&delay={delay_}"
            for  delay_, duration_, code_, numbytes_ in zip(
                (self.RNG.uniform(0, delay) for _ in range(n_trials)),
                (self.RNG.uniform(0, duration) for _ in range(n_trials)),
                (code for code in self.RNG.choices(STATUS_CODES, k=n_trials)),
                (numbyte for numbyte in self.RNG.choices(range(5, numbyte), k=n_trials))
            )
        )

    def BYTES_URLS(self, n_trials=1000, numbytes=100):
        return (
            f"{self.BASE_URL}/bytes/{numbyte}"
            for numbyte in self.RNG.choices(range(numbytes), k=n_trials)
        )

    def STATUS_CODE_URLS(self, n_trials=1000):
        return (
            f"{self.BASE_URL}/status/{code}"
            for code in self.RNG.choices(STATUS_CODES, k=n_trials)
        )

    def IMAGE_URLS(self, n_trials=1000):
        return (
            f"{self.BASE_URL}/image/{image_type}"
            for image_type in self.RNG.choices(IMAGE_TYPES, k=n_trials)
        )


from time import time
from sqlite3 import connect
from functools import wraps as wraps

from .abstract import Cache

def cache_expiration_trigger(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.options.cache_timeout:
            self._cursor.execute("DELETE FROM cache WHERE expiration < strftime('%s', 'now');")
        result = func(self, *args, **kwargs)
        self._instance._sqlite_conn.commit()
        return result
    return wrapper

class SQLiteCache(Cache):
    def __init__(
        self,
        instance,
        **kwargs
    ):
        self._instance = instance
        super().__init__(backend="sqlite", **kwargs)



    @property
    def newconn(self):
        return connect(self.options.db)


    def _create_tables(self):
        with self._instance._sqlite_conn:
            self._instance._sqlite_conn.execute("""
                CREATE TABLE IF NOT EXISTS cache (
                key TEXT PRIMARY KEY,
                value BLOB,
                expiration FLOAT
                )
            """)
            if self.options.cache_timeout:
                self._instance._sqlite_conn.execute("DELETE FROM cache WHERE expiration < strftime('%s', 'now');")

    @cache_expiration_trigger
    def __contains__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)
        self._cursor.execute("SELECT key FROM cache WHERE key = ?", (key,))
        return bool(self._cursor.fetchone())

    @Cache.deserialize
    def __getitem__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)

        self._cursor.execute("SELECT value,expiration FROM cache WHERE key = ?", (key,))
        value = self._cursor.fetchone()

        if value is not None:
            data, expiration = value
            if self.options.cache_timeout and expiration < time():
                del self[key]
                return None
            if self.options.compression:
                data = self._decompress(data)
            return data
        return None

    @Cache.serialize
    def __setitem__(self, key, value):
        if not key.endswith(":cache"):
            key = self._parse_key(key)

        if self.options.compression:
            value = self._compress(value)

        expiration = time() + self.options.cache_timeout if self.options.cache_timeout else 0

        with self._instance._sqlite_conn:
            self._instance._sqlite_conn.execute("INSERT OR REPLACE INTO cache (key, value, expiration) VALUES (?, ?, ?)", (key, value, expiration))

    @cache_expiration_trigger
    def __delitem__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)
        with self._instance._sqlite_conn:
            self._instance._sqlite_conn.execute("DELETE FROM cache WHERE key = ?", (key,))

    def clear(self):
        with self._instance._sqlite_conn:
            return self._instance._sqlite_conn.execute("DELETE FROM cache")

    @cache_expiration_trigger
    def keys(self):
        self._cursor.execute("SELECT key FROM cache")
        return tuple(key[0] for key in self._cursor.fetchall())

    @cache_expiration_trigger
    def values(self):
        self._cursor.execute("SELECT value FROM cache")
        return tuple(self._decompress(value[0]) if self.options.compression else value[0] for value in self._cursor.fetchall())

    @cache_expiration_trigger
    def items(self):
        self._cursor.execute("SELECT key, value FROM cache")
        return tuple((key, self._decompress(value)) if self.options.compression else (key, value) for key, value in self._cursor.fetchall())

    def _cleanup(self):
        pass


import time

from redislite import Redis

from .abstract import Ratelimit, RatelimitDecoratorMixin


class LeakyBucket(Ratelimit):
    __slots__ = ("_capacity", "_leak_rate")

    def __init__(
        self,
        capacity: int | float   = 10,
        leak_rate: int | float  = 5,
        key: str | None         = None,
        conn: Redis | None     = None,
        **kwargs
    ):
        self._capacity = capacity  # Maximum capacity of the bucket
        self._leak_rate = leak_rate  # Leak rate in requests per second
        super().__init__(key=key, conn=conn, backend="redis", **kwargs)


    def _leak(self, key):
        # Get the key information
        data = self._instance._redis_conn.hgetall(key) # type: ignore
        if not data: # type: ignore
            now = time.time()
            self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"content": 0, "last_check": now, "last_update": now}) # type: ignore
            return 0

        content = float(data[b"content"]) # type: ignore
        last_check = float(data[b"last_check"]) # type: ignore

        # Calculate the amount of time that has passed
        now = self.now
        elapsed = now - last_check

        # Leak the appropriate amount of requests
        content -= elapsed * self._leak_rate
        content = max(content, 0)  # Ensure content doesn't go negative

        # Store the data
        self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"content": content, "last_check": now, "last_update": now}) # type: ignore
        return content


    def ok(self, key):
        content = self._leak(key)
        if content < self._capacity:
            self._set_redis_key(key, self._instance._redis_conn.hset, "content", content + 1) # type: ignore
            return True
        return False



class TokenBucket(Ratelimit):
    __slots__ = ("_capacity", "_fill_rate")

    def __init__(
        self,
        capacity: float | int       = 20,
        fill_rate: float | int      = 5,
        key: str | int | None       = None,
        conn: Redis | None         = None,
        **kwargs
    ):
        self._capacity = float(capacity)
        self._fill_rate = float(fill_rate)
        super().__init__(key=key, conn=conn, backend="redis", **kwargs)


    def get_tokens(self, key):
        data = self._instance._redis_conn.hgetall(key) # type: ignore
        if not data:
            now = time.time()
            self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"tokens": self._capacity, "last_fill": now, "last_update": now}) # type: ignore
            return self._capacity

        tokens = float(data[b"tokens"]) # type: ignore
        last_fill = float(data[b"last_fill"]) # type: ignore

        # Calculate the time elapsed since the last fill
        now = time.time()
        elapsed = now - last_fill

        # Calculate the number of tokens to add based on the fill rate
        to_add = elapsed * (self._fill_rate / self._capacity)

        # Set the new number of tokens (up to the capacity)
        tokens = min(self._capacity, tokens + to_add)

        # Store the new number of tokens and the last fill time
        if tokens < 1:
            self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"tokens": tokens, "last_fill": now, "last_update": now}) # type: ignore
            return False

        tokens -= 1
        self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"tokens": tokens, "last_fill": now, "last_update": now}) # type: ignore
        return True


    def ok(self, key):
        return self.get_tokens(key)



class SlidingWindow(Ratelimit):
    __slots__ = ("_limit", "_window")

    def __init__(
        self,
        limit: int              = 10,
        window: float | int     = 1.0,
        key: str | None         = None,
        conn: Redis | None     = None,
        **kwargs
    ):
        self._limit = limit
        self._window = window * 1000000000
        super().__init__(key=key, conn=conn, backend="redis", **kwargs)


    @property
    def limit(self):
        return self._limit


    @property
    def window(self):
        return self._window


    @property
    def edge(self):
        return self.current_timestampns - self.window


    def ok(self, key):
        self._set_redis_key(key, self._instance._redis_conn.zremrangebyscore, 0, self.edge) # type: ignore
        count = self._instance._redis_conn.zcard(key) # type: ignore
        if count < self.limit: # type: ignore
            ts = self.current_timestampns
            self._set_redis_key(key, self._instance._redis_conn.zadd, mapping={ts:ts})# type: ignore
            return True
        return False


class FixedWindow(Ratelimit):
    __slots__= ("_limit", "_window")

    def __init__(
        self,
        limit: int              = 10,
        window: float | int     = 1.0,
        key: str | None         = None,
        conn: Redis | None     = None,
        **kwargs
    ):
        self._limit = limit
        self._window = window
        super().__init__(key=key, conn=conn, backend="redis", **kwargs)


    def ok(self, key):
        data = self._instance._redis_conn.hgetall(key) # type: ignore
        if not data:
            now = time.time()
            self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"requests": 1, "window_start": now, "last_update": now}) # type: ignore

            return True

        window_start = float(data[b"window_start"]) # type: ignore
        requests = int(data[b"requests"]) # type: ignore
        current_time = time.time()
        if current_time - window_start > self._window:
            requests = 0
            self._instance._redis_conn.hset(key, mapping={"requests": requests, "window_start": time.time(), "last_update": current_time}) # type: ignore

        if requests < self._limit:
            self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"requests": requests + 1, "last_update": current_time}) # type: ignore
            return True

        self._set_redis_key(key, self._instance._redis_conn.hset, "last_update", current_time) # type: ignore
        return False


class GCRA(Ratelimit):
    __slots__ = ("_period", "_limit")

    def __init__(
        self,
        period: float | int     = 1,
        limit: int              = 10,
        key: str | None         = None,
        conn: Redis | None     = None,
        **kwargs
    ):
        self._period = period
        self._limit = limit
        super().__init__(key=key, conn=conn, backend="redis", **kwargs)


    def ok(self, key):
        data = self._instance._redis_conn.hgetall(key) # type: ignore
        if not data:
            now = time.time()
            self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"last_time": now, "last_update": now}) # type: ignore
            return True

        last_time = float(data[b"last_time"]) # type: ignore
        current_time = time.time()
        expected_time = last_time + self._period

        if current_time < expected_time - self._limit:
            self._set_redis_key(key, self._instance._redis_conn.hset, "last_update", current_time) # type: ignore
            return False
        else:
            self._set_redis_key(key, self._instance._redis_conn.hset, mapping={"last_time": max(expected_time, current_time), "last_update": current_time}) # type: ignore
            return True


_CLASS_TYPES = {
    "leakybucket": LeakyBucket,
    "tokenbucket": TokenBucket,
    "slidingwindow": SlidingWindow,
    "fixedwindow": FixedWindow,
    "gcra": GCRA,
}

class RedisRatelimitFactory:
    def __new__(cls, instance, type="slidingwindow", key=None, *args, **kwargs):
        type_name = type
        type = _CLASS_TYPES[type_name]
        self = type.__new__(type)
        self._instance = instance
        self.__init__(*args, key=key, **kwargs)
        self._ratelimit_type = type_name
        return self


class RedisRatelimitDecoratorFactory:
    _type = type

    def __new__(cls, type="slidingwindow", *args, **kwargs):
        type_name = type
        new_type = cls._type(type, (_CLASS_TYPES[type_name], RatelimitDecoratorMixin), {})
        self = new_type.__new__(new_type)
        self.__init__(*args, **kwargs)
        return self


import traceback
import asyncio
from time import perf_counter as _perf_counter
from datetime import timedelta as _timedelta
from itertools import tee

from orjson import dumps as _jsondump
from alive_progress import alive_bar as _alive_bar
from httpx import AsyncClient as _AsyncClient
from aiohttp import (
    ClientSession as _ClientSession,
    ClientTimeout as AIOHTTPTimeout,
    ClientError as _ClientError,
)

from .meta import SessionMeta as _SessionMeta
from .utils import get_valid_kwargs as _get_valid_kwargs, take as _take
from .useragents import useragent
from .objects import Response as _Response, Timeouts
from .config import SessionConfig as config


def _dumps(obj):
    return _jsondump(obj).decode("utf-8")


class AsyncSession(metaclass=_SessionMeta):
    """
    An asynchronous session for making HTTP requests.

    Args:
        headers (dict, optional): Custom headers to be included in each request. Defaults to None.
        json_serialize (callable, optional): Custom JSON serialization function. Defaults to _dumps.
        random_user_agents (bool, optional): Whether to include random user agents in the request headers. Defaults to True.
        **kwargs: Additional keyword arguments to be passed to the underlying _ClientSession.

    Attributes:
        _random_user_agents (bool): Whether random user agents are enabled.
        _session (ClientSession): The underlying client session.
        _semaphore (Semaphore): A semaphore used for limiting the number of concurrent requests.

    Methods:
        __aenter__(): Asynchronous context manager entry point.
        __aexit__(exc_type, exc_value, exc_traceback): Asynchronous context manager exit point.
        retrieve_response(resp): Asynchronously retrieve the response from a request.
        request(url, method, headers=None, *, bar=None, **kwargs): Asynchronously send a request.
        requests(urls, method="GET", *, headers=None, progress=True, **kwargs): Asynchronously send multiple requests.
        get(url, **kwargs): Asynchronously send a GET request.
        head(url, **kwargs): Asynchronously send a HEAD request.
        options(url, **kwargs): Asynchronously send an OPTIONS request.
        delete(url, **kwargs): Asynchronously send a DELETE request.
        post(url, **kwargs): Asynchronously send a POST request.
        put(url, **kwargs): Asynchronously send a PUT request.
        patch(url, **kwargs): Asynchronously send a PATCH request.
    """
    __slots__ = ("_random_user_agents", "_session", "_semaphore", "_limit")

    def __init__(
        self,
        headers: dict | None                                = None,
        json_serialize: callable                            = _dumps,
        random_user_agents: bool                            = True,
        **kwargs
    ):
        if (loop := kwargs.pop("loop", None)) is None:
            try:
                loop = asyncio.get_event_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()

        headers = headers if isinstance(headers, dict) else {}
        kwargs["loop"] = loop
        kwargs["raise_errors"] = kwargs.pop("raise_errors", None) or config.raise_errors
        kwargs = self._parse_kwargs(kwargs)
        kwargs = _get_valid_kwargs(_ClientSession.__init__, kwargs)
        self._session = _ClientSession(headers=headers, json_serialize=json_serialize, **kwargs)
        self._random_user_agents = random_user_agents
        self._limit = max(self._session.connector.limit, 100) if self._session.connector.limit > 0 else 1
        if config.semaphore == "global":
            AsyncSession._semaphore = asyncio.Semaphore(self._limit)
        else:
            self._semaphore = asyncio.Semaphore(self._limit)


    async def __aenter__(self):
        return self


    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        if exc_traceback:
            import traceback
            traceback.print_exception(exc_type, exc_value, exc_traceback)
        await self._session.close()


    def _parse_kwargs(self, kwargs):
        keys = set(kwargs.keys())
        timeouts = set(_take(lambda x: x in Timeouts.AIOHTTP.value, keys))

        if timeouts:
            args = {"total": kwargs.pop("timeout", None)}
            if  "pool_timeout" in timeouts and (pool := kwargs.pop("pool_timeout", "None")) != "None" or\
                "connect_timeout" in timeouts and (pool := kwargs.pop("connect_timeout", "None")) != "None":
                args["connect"] = connect

            if "connect_timeout" in timeouts and (connect := kwargs.pop("connect_timeout", "None")) != "None" or\
               "sock_connect" in timeouts and (connect := kwargs.pop("sock_connect", "None")) != "None":
                args["connect"] = connect

            if "read_timeout" in timeouts and (read := kwargs.pop("read_timeout", "None")) != "None" or\
               "sock_read" in timeouts and (read := kwargs.pop("sock_read", "None")) != "None":
                args["read"] = read

            kwargs["timeout"] = AIOHTTPTimeout(**args)
        return kwargs


    async def _retrieve_response(self, resp, callbacks, bar):
        response = _Response(**{
            "version": resp.version,
            "status": resp.status,
            "ok": resp.ok,
            "method": resp.method,
            "reason": resp.reason,
            "url": resp.url,
            "real_url": resp.real_url,
            "connection": resp.connection,
            "content": await resp.content.read() if not isinstance(resp.content, bytes) else resp.content,
            "cookies": resp.cookies,
            "headers": resp.headers,
            "raw_headers": resp.raw_headers,
            "links": resp.links,
            "content_type": resp.content_type,
            "charset": resp.charset,
            "content_disposition": resp.content_disposition,
            "history": resp.history,
            "request": resp.request_info,
            "elapsed": resp.elapsed,
        })
        return self._run_callbacks(response, callbacks, bar)


    async def request(self, url, method, headers=None, *, bar=None, callbacks=None, **kwargs):
        if self._random_user_agents:
            headers = headers or {}
            headers["User-Agent"] = useragent()

        kwargs = _get_valid_kwargs(_ClientSession._request, kwargs)
        begin = _perf_counter()

        try:
            async with self._session.request(method=method, url=url, headers=headers, **kwargs) as response:
                end = _perf_counter()
                response.elapsed = _timedelta(seconds=end - begin)
                return await self._retrieve_response(response, callbacks, bar)

        except asyncio.TimeoutError as e:
            if config.raise_errors:
                raise e
            response =_Response(status=408, errors=e, reason="TimeoutError", content=str(e))
            return self._run_callbacks(response, callbacks, bar)


        except _ClientError as e:
            if config.raise_errors:
                raise e
            if hasattr(response, "status") and response.status is not None:
                status = response.status
            else:
                status = 500
            response = _Response(status=status, errors=e, reason="ClientError", content=str(e))
            return self._run_callbacks(response, callbacks, bar)

        except Exception as e:
            if config.raise_errors:
                raise e
            response = _Response(status=500, errors=e, reason="Exception", content=str(e))
            return self._run_callbacks(response, callbacks, bar)





    # NEEDS FIXED
    async def requests(self, urls, method="GET", *, headers=None, progress=True, ratelimit=True, use_cache=True, callbacks=None, **kwargs):
        if not isinstance(method, str) and isinstance(method, (list, tuple)):
            items = zip(urls, method)
        else:
            items = ((url, method) for url in urls)

        if progress:
            items, items_ = tee(items)
            item_length = sum(1 for _ in items_)
            with _alive_bar(item_length) as bar:
                results = await asyncio.gather(*(self.request(url=url, method=method, headers=headers, ratelimit=ratelimit, use_cache=use_cache, bar=bar, callbacks=callbacks, **kwargs) for url, method in items))
        else:
            results = await asyncio.gather(*(self.request(url=url, method=method, headers=headers, ratelimit=ratelimit, use_cache=use_cache, callbacks=callbacks, **kwargs) for url, method in items))
        return results


    async def get(self, url, **kwargs):
        return await self.request(url=url, method="GET", **kwargs)


    async def head(self, url, **kwargs):
        return await self.request(url=url, method="HEAD", **kwargs)


    async def options(self, url, **kwargs):
        return await self.request(url=url, method="OPTIONS", **kwargs)


    async def delete(self, url, **kwargs):
        return await self.request(url=url, method="DELETE", **kwargs)


    async def post(self, url, **kwargs):
        return await super().request(url=url, method="POST", **kwargs)


    async def put(self, url, **kwargs):
        return await super().request(url=url, method="PUT", **kwargs)


    async def patch(self, url, **kwargs):
        return await self.request(url=url, method="PATCH", **kwargs)



class AsyncClient(_AsyncClient):
    def __init__(self, headers=None, http2=True, random_user_agents=True, **kwargs):
        headers = headers if isinstance(headers, dict) else {}
        self._random_user_agents = random_user_agents
        session_kwargs = _get_valid_kwargs(_AsyncClient.__init__, kwargs)
        super().__init__(headers=headers, http2=http2, **session_kwargs)
        self._headers = headers


    async def __aenter__(self):
        return self


    async def __aexit__(self, *args):
        await self.aclose()


    async def request(self, method, url, *, headers=None, **kwargs):
        if self._random_user_agents:
            headers = headers or {}
            headers["User-Agent"] = useragent()
        return await super().request(method, url, headers=headers, **kwargs)


import sqlite3
from abc import ABC, abstractmethod
from atexit import register
from functools import wraps
from zlib import compress, decompress

import orjson as json
from redislite import Redis

from ..objects import Response, CacheOptions, CacheInMemoryCache


class Cache(ABC):
    """
    Abstract base class for cache implementations.

    Attributes:
        options (CacheOptions): The cache options.
        _backend (str): The backend used for caching.
        _conn (InMemoryCacheObject | Redis | sqlite3.Connection): The cache object.
        _key (str): The cache key.
    """

    __slots__ = ("options", "_instance", "_backend", "_key")

    def __init__(
        self,
        backend: str | None             = None,
        key: str | None                 = None,
        conn: Redis | None             = None,
        cache_timeout: float | int      = 3600,
        check_frequency: float | int    = 30,
        **kwargs,
    ):
        self._backend = backend
        self._key = key if isinstance(key, str) else "Session"

        kwargs = {
            "cache_timeout": cache_timeout,
            "check_frequency": check_frequency,
            **kwargs
        }

        if kwargs.get("cache_options") is not None:
            kwargs = {**kwargs, **kwargs["cache_options"]}
        elif kwargs.get("cache") is not None:
            kwargs = {**kwargs, **kwargs["cache"]}

        self.options = CacheOptions.from_backend(backend, **kwargs)

        if self._backend == "memory":
            if not hasattr(self._instance, "_memory_conn"):
                self._instance.__class__._cache_memory_conn = CacheInMemoryCache(options=self.options) # type: ignore

        elif self._backend == "redis":
            if not hasattr(self._instance, "_redis_conn") or not isinstance(self._instance._redis_conn, Redis):
                self._instance.__class__._redis_conn = conn or Redis(**self.options.redis_server_config())

        elif self._backend == "sqlite":
            assert self.options.db is not None, "Database file must be provided for SQLite cache."
            if not hasattr(self._instance, "_sqlite_conn") or not isinstance(self._instance._sqlite_conn, sqlite3.Connection):
                self._instance.__class__._sqlite_conn = conn or sqlite3.connect(self.options.db)
            if not hasattr(self, "_cursor"):
                self._cursor = self._instance._sqlite_conn.cursor()
            self._create_tables()

        register(self._cleanup)


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} cached={len(self.keys())}>"

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} cached={len(self.keys())}>"


    @property
    def cache(self):
        """
        Get the cache object based on the backend.

        Returns:
            CacheInMemoryCache | Redis | sqlite3.Connection: The cache object.
        """
        if self._backend == "memory":
            return self._instance._cache_memory_conn
        elif self._backend == "redis":
            return self._instance._redis_conn
        elif self._backend == "sqlite":
            return self._instance._sqlite_conn

    @abstractmethod
    def __contains__(self, key):
        pass

    @abstractmethod
    def __getitem__(self, key):
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        pass

    @abstractmethod
    def __delitem__(self, key):
        pass

    @abstractmethod
    def _cleanup(self):
        pass

    @abstractmethod
    def keys(self):
        pass

    @abstractmethod
    def values(self):
        pass

    @abstractmethod
    def items(self):
        pass

    @staticmethod
    def serialize(func):
        @wraps(func)
        def wrapper(self, key, response):
            response = json.dumps(response.serialize())
            value = func(self, key, response)
            return value
        return wrapper

    @staticmethod
    def deserialize(func):
        @wraps(func)
        def wrapper(self, key):
            response = func(self, key)
            return Response.deserialize(json.loads(response)) if response is not None else response
        return wrapper

    def _compress(self, value):
        return compress(value)

    def _decompress(self, value):
        return decompress(value)

    def clear_cache(self):
        """
        Clear the cache.
        """
        self.cache.clear()

    def _parse_key(self, url):
        """
        Parse the cache key.

        Args:
            url (str): The URL to parse.

        Returns:
            str: The parsed cache key.
        """
        return ":".join((self._key, url, "cache"))


import traceback
import asyncio
from concurrent.futures import as_completed
from itertools import tee
from atexit import register as _register
from typing import (
    List,
    Tuple,
    Set,
    Generator,
)

from requests import Session as _Session
from alive_progress import alive_bar as _alive_bar
from httpx._exceptions import TimeoutException as _TimeoutException
from httpx import (
    Client as _HTTPXSession,
    Timeout as _HTTPXTimeout,
    Limits as _HTTPXLimits,
)

from .asyncsession import AsyncSession as _AsyncSession
from .thread import AsyncLoopThread
from .config import SessionConfig as config
from .meta import SessionMeta as _SessionMeta
from .useragents import useragent
from .objects import (
    Response,
    Timeouts as _Timeouts
)
from .utils import (
    take as _take,
    get_valid_kwargs as _get_valid_kwargs,
)



class Session(_HTTPXSession, metaclass=_SessionMeta):
    _ID = 0

    def __init__(
        self,
        headers: dict                           = None,
        http2: bool                             = True,
        random_user_agents: bool                = True,
        threaded: bool                          = True,
        **kwargs
    ):
        """
        Initializes a Session object.

        Args:
            headers (dict, optional): Custom headers to be included in the requests. Defaults to None.
            http2 (bool, optional): Flag indicating whether to use HTTP/2. Defaults to True.
            random_user_agents (bool, optional): Flag indicating whether to use random user agents. Defaults to True.
            **kwargs: Additional keyword arguments to be passed to the _HTTPXSession constructor.
        """
        Session._ID += 1
        self._id = Session._ID

        self._threaded = threaded
        self._raise_errors = kwargs.pop("raise_errors", None) or config.raise_errors
        self._headers = headers if isinstance(headers, dict) else {}
        self._random_user_agents = random_user_agents

        if self._threaded:
            kwargs["timeout"] = kwargs.pop("timeout", None) or config.threaded_timeout
            self._loop = asyncio.new_event_loop()
            self._thread = AsyncLoopThread(target=self.__start_event_loop, name=f"Session {self._id}-EventLoopThread", daemon=True)
            self._thread.start()

        self.__kwargs = kwargs.copy()
        kwargs = self._parse_kwargs(kwargs)

        kwargs = _get_valid_kwargs(_HTTPXSession.__init__, kwargs)
        super().__init__(http2=http2, headers=self._headers, **kwargs)


    def __call__(self):
        self.__kwargs["threaded"] = False
        return self.__class__(**self.__kwargs)

    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_traceback:
            import traceback
            traceback.print_exception(exc_type, exc_value, exc_traceback)
        self._cleanup()


    def _cleanup(self):
        if self._threaded and hasattr(self, "_loop"):
            self._thread.stop()
            self._loop.close()
        self.close()


    def _parse_kwargs(self, kwargs):
        keys = set(kwargs.keys())
        timeouts = set(_take(lambda x: x in _Timeouts.HTTPX.value, keys))
        limits = set(_take(lambda x: x in {"max_keepalive", "max_connections", "max_concurrent_requests", "limits"}, keys))

        if timeouts:
            if "timeout" in timeouts and isinstance(kwargs["timeout"], _HTTPXTimeout):
                pass
            else:
                args = {}
                if "connect_timeout" in timeouts and (connect := kwargs.pop("connect_timeout", "None")) != "None":
                    args["connect"] = connect

                if "read_timeout" in timeouts and (read := kwargs.pop("read_timeout", "None")) != "None":
                    args["read"] = read

                if "write_timeout" in timeouts and (write := kwargs.pop("write_timeout", "None")) != "None":
                    args["write"] = write

                if "pool_timeout" in timeouts and (pool := kwargs.pop("pool_timeout", "None")) != "None":
                    args["pool"] = pool

                kwargs["timeout"] = _HTTPXTimeout(
                    timeout=kwargs.pop("timeout", 5),
                    **args
                )

        if limits:
            if "limits" in limits:
                if isinstance(kwargs["limits"], _HTTPXLimits):
                    pass
                else:

                    kwargs["limits"] = _HTTPXLimits(
                        max_keepalive_connections       = kwargs.pop("max_keepalive_connections", None),
                        max_connections                 = kwargs.pop("max_connections", None),
                        keepalive_expiry                = kwargs.pop("keepalive_expiry", 5),
                    )
        return kwargs


    def _retrieve_response(self, resp, callbacks, bar):
        response = Response(**{
            "version":  resp.http_version,
            "status":   resp.status_code,
            "ok":       resp.status_code < 400,
            "method":   resp.request.method,
            "reason":   resp.reason_phrase,
            "url":      resp.url,
            "content":  resp.content,
            "encoding": resp.encoding,
            "cookies":  resp.cookies,
            "headers":  resp.headers,
            "history":  resp.history,
            "request":  resp.request,
            "elapsed":  resp.elapsed,
        })
        return self._run_callbacks(response, callbacks, bar)


    def request(self, method, url, *, headers=None, threaded=False, bar=None, callbacks=None, **kwargs):
        """
        Sends an HTTP request.

        Args:
            method (str): The HTTP method to be used.
            url (str): The URL to send the request to.
            headers (dict, optional): Custom headers to be included in the request. Defaults to None.
            bar (callable, optional): A progress bar function. Defaults to None.
            **kwargs: Additional keyword arguments to be passed to the super().request method.

        Returns:
            The response from the HTTP request.
        """
        if self._random_user_agents:
            headers = headers if isinstance(headers, dict) else {}
            headers["User-Agent"] = useragent()

        kwargs = _get_valid_kwargs(super().request, kwargs)
        try:
            response = super().request(method=method, url=url, headers=headers, **kwargs)
            return self._retrieve_response(response, callbacks, bar)
        except _TimeoutException as e:
            if self._raise_errors:
                raise e
            return Response(status=408, ok=False, reason="Request Timeout", url=url, errors=e)
        except Exception as e:
            if self._raise_errors:
                raise e
            return Response(status=500, ok=False, reason="Internal Server Error", url=url, errors=e)

    def __async_session(self):
        class AsyncSession(*(base for name, base in self.__bases__.items() if name.endswith("Mixin")), _AsyncSession):
            pass
        return AsyncSession(loop=self._loop, **self.__kwargs)

    def __start_event_loop(self):
        asyncio.set_event_loop(self._loop)
        return self._loop

    async def __requests(self, urls, method="GET", headers=None, progress=None, callbacks=None, **kwargs):
        async with self.__async_session() as session:
            return await session.requests(urls=urls, method=method, headers=headers, progress=progress, callbacks=callbacks, **kwargs)

    def requests(
        self,
        urls:   str | List[str] | Tuple[str, ...] | Set[str] | Generator[str, None, None],
        method: str | List[str] | Tuple[str, ...] | Set[str] | Generator[str, None, None]       = "GET",
        headers: dict | None                                                                    = None,
        progress: bool                                                                          = True,
        threaded: bool                                                                          = True,
        callbacks: List[callable] | Tuple[callable, ...] | Set[callable]                        = None,
        **kwargs
    ):
            """
            Sends multiple HTTP requests asynchronously.

            Args:
                urls (list or tuple): List of URLs to send requests to.
                method (str or list or tuple, optional): HTTP method to use for the requests. Defaults to "GET".
                headers (dict, optional): Headers to include in the requests. Defaults to None.
                progress (bool, optional): Whether to display progress bar. Defaults to True.
                threaded (bool, optional): Whether to use threaded execution. Defaults to True.
                callbacks (list or tuple, optional): List of callback functions to execute after each request. Defaults to None.
                **kwargs: Additional keyword arguments to pass to the requests.

            Returns:
                tuple: A tuple of the results from each request.

            """
            if self._threaded and threaded:
                future = asyncio.run_coroutine_threadsafe(self.__requests(urls=urls, method=method, headers=headers, progress=progress, callbacks=callbacks, **kwargs), self._loop)
                return tuple(future.result() for future in as_completed((future,)))[0]

            if isinstance(urls, str):
                urls = (urls,)

            if not isinstance(method, str) and isinstance(method, (list, tuple)):
                items = zip(urls, method)
            else:
                items = ((url, method) for url in urls)

            if progress:
                items, items_ = tee(items)
                item_length = sum(1 for _ in items_)

                with _alive_bar(item_length) as bar:
                        results = tuple(self.request(method=method, url=url, headers=headers, callbacks=callbacks, bar=bar, **kwargs) for url, method in items)
            else:
                results = tuple(self.request(method=method, url=url, headers=headers, callbacks=callbacks, **kwargs) for url, method in items)

            return results


    def get(self, url, **kwargs):
        """
        Sends an HTTP GET request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="GET", url=url, **kwargs)


    def head(self, url, **kwargs):
        """
        Sends an HTTP HEAD request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="HEAD", url=url, **kwargs)


    def options(self, url, **kwargs):
        """
        Sends an HTTP OPTIONS request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="OPTIONS", url=url, **kwargs)


    def delete(self, url, **kwargs):
        """
        Sends an HTTP DELETE request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="DELETE", url=url, **kwargs)


    def post(self, url, **kwargs):
        """
        Sends an HTTP POST request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="POST", url=url, **kwargs)


    def put(self, url, **kwargs):
        """
        Sends an HTTP PUT request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="PUT", url=url, **kwargs)


    def patch(self, url, **kwargs):
        """
        Sends an HTTP PATCH request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="PATCH", url=url, **kwargs)


class RequestsSession(_Session, metaclass=_SessionMeta):
    def __init__(
        self,
        headers=None,
        random_user_agents=True,
        threaded=False,
        **kwargs
    ):
        """
        Initializes a Session object.

        Args:
            headers (dict, optional): Custom headers to be included in the requests. Defaults to None.
            http2 (bool, optional): Flag indicating whether to use HTTP/2. Defaults to True.
            random_user_agents (bool, optional): Flag indicating whether to use random user agents. Defaults to True.
            **kwargs: Additional keyword arguments to be passed to the _HTTPXSession constructor.
        """
        super().__init__()
        self._headers = headers if isinstance(headers, dict) else {}
        self._random_user_agents = random_user_agents
        kwargs = _get_valid_kwargs(_HTTPXSession.__init__, kwargs)

        self._threaded = threaded
        if self._threaded:
            kwargs["timeout"] = config.threaded_timeout
            self._loop = asyncio.new_event_loop()
            self._thread = AsyncLoopThread(target=self.__start_event_loop, name=f"Session {self._id}-EventLoopThread", daemon=True)
            self._thread.start()
            _register(self._cleanup)


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_traceback:
            import traceback
            traceback.print_exception(exc_type, exc_value, exc_traceback)
        self._cleanup()
        self.close()

    def _cleanup(self):
        if self._threaded and hasattr(self, "_loop"):
            self._thread.stop()
            self._loop.close()


    @staticmethod
    def _retrieve_response(resp):
        return Response(**{
            "status": resp.status_code,
            "ok": resp.ok,
            "method": resp.request.method,
            "reason": resp.reason,
            "url": resp.url,
            "content": resp.content,
            "encoding": resp.encoding,
            "cookies": resp.cookies,
            "headers": resp.headers,
            "links": resp.links,
            "history": resp.history,
            "request": resp.request,
            "elapsed": resp.elapsed,
        })

    def request(self, method, url, *, headers=None, threaded=False, bar=None, **kwargs):
        """
        Sends an HTTP request.

        Args:
            method (str): The HTTP method to be used.
            url (str): The URL to send the request to.
            headers (dict, optional): Custom headers to be included in the request. Defaults to None.
            bar (callable, optional): A progress bar function. Defaults to None.
            **kwargs: Additional keyword arguments to be passed to the super().request method.

        Returns:
            The response from the HTTP request.
        """
        if self._threaded and threaded:
            return self._request(method, url, headers=headers, bar=bar, **kwargs)

        if self._random_user_agents:
            headers = headers if isinstance(headers, dict) else {}
            headers["User-Agent"] = _UserAgents.user_agent
        kwargs = _get_valid_kwargs(super().request, kwargs)
        results = super().request(method=method, url=url, headers=headers, **kwargs)
        if bar is not None:
            bar()
        return self._retrieve_response(results)


    def requests(self, urls, method="GET", *, headers=None, threaded=True, progress=True, **kwargs):
        """
        Sends multiple HTTP requests.

        Args:
            urls (iterable): An iterable of URLs to send the requests to.
            method (str, optional): The HTTP method to be used. Defaults to "GET".
            headers (dict, optional): Custom headers to be included in the requests. Defaults to None.
            progress (bool, optional): Flag indicating whether to show a progress bar. Defaults to False.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            A tuple of responses from the HTTP requests.
        """

        if not isinstance(method, str) and isinstance(method, (list, tuple)):
            items = zip(urls, method)
        else:
            items = ((url, method) for url in urls)

        if progress:
            items, items_ = tee(items)
            item_length = sum(1 for _ in items_)
            with _alive_bar(item_length) as bar:
                if self._threaded and threaded:
                    futures = tuple(self._request(method=method, url=url, headers=headers, bar=bar, **kwargs) for url, method in items)
                    results = tuple(future.result() for future in as_completed(futures))
                else:
                    results = tuple(self.request(method=method, url=url, headers=headers, bar=bar, **kwargs) for url, method in items)
        else:
            if self._threaded and threaded:
                futures = tuple(self._request(method=method, url=url, headers=headers, **kwargs) for url, method in items)
                results = tuple(future.result() for future in as_completed(futures))
            else:
                results = tuple(self.request(method=method, url=url, headers=headers, **kwargs) for url, method in items)
        return results


    def get(self, url, **kwargs):
        """
        Sends an HTTP GET request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="GET", url=url, **kwargs)


    def head(self, url, **kwargs):
        """
        Sends an HTTP HEAD request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="HEAD", url=url, **kwargs)


    def options(self, url, **kwargs):
        """
        Sends an HTTP OPTIONS request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="OPTIONS", url=url, **kwargs)


    def delete(self, url, **kwargs):
        """
        Sends an HTTP DELETE request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="DELETE", url=url, **kwargs)


    def post(self, url, **kwargs):
        """
        Sends an HTTP POST request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="POST", url=url, **kwargs)


    def put(self, url, **kwargs):
        """
        Sends an HTTP PUT request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="PUT", url=url, **kwargs)


    def patch(self, url, **kwargs):
        """
        Sends an HTTP PATCH request.

        Args:
            url (str): The URL to send the request to.
            **kwargs: Additional keyword arguments to be passed to the request method.

        Returns:
            The response from the HTTP request.
        """
        return self.request(method="PATCH", url=url, **kwargs)


import traceback
from functools import wraps
from inspect import iscoroutinefunction
from asyncio import iscoroutinefunction as _iscoroutinefunction

from .objects import Response
from .config import SessionConfig as config


def clear_cache(self):
    if hasattr("self", "_cache"):
        self._cache.clear()
    if hasattr(self, "_ratelimiter"):
        self._ratelimiter.clear()


def _run_callbacks(
        self,
        response:   Response,
        callbacks:  tuple | list | set,
        bar:        callable
    ):
    if callbacks is None or response.errors is not None and not config.run_callbacks_on_error:
        if bar is not None:
            bar()
        return response

    rets = tuple()
    for callback in callbacks:
        try:
            if callable(callback):
                ret = callback(response)
                rets = (*rets, ret)
            else:
                if config.print_callback_exceptions:
                    print(f"Callback {callback} is not callable.")
        except Exception as e:
            rets = (*rets, e)
            if config.print_callback_exceptions:
                print(f"Callback {callback} raised exception: {e}")
            if config.print_callback_tracebacks:
                traceback.print_exception(e)

    if config.return_callbacks:
        response.set_callbacks(rets)

    if bar is not None:
        bar()

    return response

def cache(func):
    @wraps(func)
    def request(self, url, method, *, headers=None, callbacks=None, cache=True, ratelimit=True, keys=None, bar=None, threaded=False, **kwargs):
        if cache:
            response = self._cache[url]
            if response is not None:
                return self._retrieve_response(response, callbacks, bar)

        response = func(self, url, method, headers=headers, callbacks=callbacks, ratelimit=ratelimit, keys=keys, bar=bar, threaded=threaded, **kwargs)

        if cache and str(response.status_code).startswith("2"):
            self._cache[url] = response

        return response

    return request

def async_cache(func):
    @wraps(func)
    async def request(self, url, method, *, headers=None, callbacks=None, cache=True, ratelimit=True, keys=None, bar=None, **kwargs):
        if cache:
            async with self._semaphore:
                response = self._cache[url]
                if response is not None:
                    return await self._retrieve_response(response, callbacks, bar)

                response = await func(self, url=url, method=method, headers=headers, callbacks=callbacks, cache=cache, ratelimit=ratelimit, keys=keys, bar=bar, **kwargs)

                if str(response.status_code).startswith("2"):
                    self._cache[url] = response

                return response
        return await func(self, url=url, method=method, headers=headers, callbacks=callbacks, cache=cache, ratelimit=ratelimit, keys=keys, bar=bar, **kwargs)
    return request


class SessionMeta(type):
    """Metaclass for session classes.

    This metaclass is responsible for dynamically modifying the session class based on its parents.
    It adds additional functionality to the session class based on the mixins present in its inheritance hierarchy.
    """

    def __new__(cls, name, bases, namespace):
        parents = {base.__name__: base for base in bases}

        if namespace.get("request") is None:
            session = None
            for name, base in parents.items():
                name = name.lower()
                if "session" in name or "client" in name:
                    session = base
                    base_index = bases.index(session)
                    if base_index != len(bases) - 1:
                        bases = list(bases)
                        bases.remove(session)
                        bases.append(session)
                        bases = tuple(bases)
                        parents = {base.__name__: base for base in bases}
                    break

            if session is not None:
                namespace["clear_cache"] = clear_cache
                namespace["_run_callbacks"] = _run_callbacks
                if hasattr(session, "request"):
                    namespace["request"] = cls.define_request(session, set(parents.keys()))

        namespace["__bases__"] = parents
        return super().__new__(cls, name, bases, namespace)


    @staticmethod
    def define_request(session, parents):
        """Define the request method based on the mixins present in the inheritance hierarchy.

        Args:
            name (str): The name of the session class.
            session (class): The base session class.
            parents (set): Set of parent class names.

        Returns:
            function: The defined request method.
        """
        if iscoroutinefunction(session.request) or _iscoroutinefunction(session.request):
            if "CacheMixin" in parents and "RatelimitMixin" in parents:
                @async_cache
                async def request(self, url, method, *, headers=None, callbacks=None, cache=True, ratelimit=True, keys=None, bar=None, **kwargs):
                    if ratelimit:
                        if not cache:
                            async with self._semaphore:
                                await self._ratelimiter.increment_async(url=url, method=method, keys=keys)
                        else:
                            await self._ratelimiter.increment_async(url=url, method=method, keys=keys)
                    return await session.request(self, url=url, method=method, headers=headers, callbacks=callbacks, bar=bar, **kwargs)

            elif "RatelimitMixin" in parents:
                async def request(self, url, method, *, headers=None, callbacks=None, ratelimit=True, keys=None, bar=None, **kwargs):
                    if ratelimit:
                        async with self._semaphore:
                            await self._ratelimiter.increment_async(url=url, method=method, keys=keys)
                    return await session.request(self, method=method, url=url, headers=headers, callbacks=callbacks, bar=bar, **kwargs)

            elif "CacheMixin" in parents:
                @async_cache
                async def request(self, url, method, *, headers=None, callbacks=None, cache=True, bar=None, **kwargs):
                    return await session.request(self, method=method, url=url, headers=headers, callbacks=callbacks, bar=bar, **kwargs)

            else:
                request = session.request

        else:
            if "CacheMixin" in parents and "RatelimitMixin" in parents:
                @cache
                def request(self, url, method, *, headers=None, threaded=False, callbacks=None, cache=True, ratelimit=True, keys=None, bar=None, **kwargs):
                    if ratelimit:
                        self._ratelimiter.increment(url=url, method=method, keys=keys)
                    return session.request(self, method=method, url=url, headers=headers, threaded=threaded, callbacks=callbacks, bar=bar, **kwargs)


            elif "RatelimitMixin" in parents:
                def request(self, url, method, *, headers=None, threaded=False, callbacks=None, ratelimit=True, keys=None, bar=None, **kwargs):
                    if ratelimit:
                        self._ratelimiter.increment(url=url, method=method, keys=keys)
                    return session.request(self, method=method, url=url, headers=headers, threaded=threaded, callbacks=callbacks, bar=bar, **kwargs)


            elif "CacheMixin" in parents:
                @cache
                def request(self, url, method, *, headers=None, threaded=False, callbacks=None, cache=True, bar=None, **kwargs):
                    return session.request(self, method=method, url=url, headers=headers, threaded=threaded, callbacks=callbacks, bar=bar, **kwargs)

            else:
                request = session.request

        return request


__all__ = "CacheMixin", "cache_factory"

from inspect import signature

from .pycache import InMemoryCache
from .redis import RedisCache
from .sqlite import SQLiteCache
from ..utils import get_valid_kwargs
from ..objects import Alias


def cache_factory(instance, backend="memory", *args, **kwargs):
    if backend in Alias.MEMORY.value:
        self = InMemoryCache(instance, *args, **kwargs)
    elif backend in Alias.REDIS.value:
        self = RedisCache(instance, *args, **kwargs)
    elif backend in Alias.SQLITE.value:
        self = SQLiteCache(instance, *args, **kwargs)
    else:
        raise NotImplementedError(f"Cache backend {backend} is not implemented.")

    return self


class CacheMixin():
    def __init__(
        self,
        backend: str                    = "memory",
        **kwargs
    ):
        self._backend = backend
        self._cache = cache_factory(self, backend=backend, **kwargs)
        kwargs["backend"] = backend

        if "kwargs" in signature(super().__init__).parameters:
            super().__init__(**kwargs)
        else:
            super().__init__(**get_valid_kwargs(super().__init__, kwargs))

    @property
    def cache(self):
        return self._cache

    @property
    def backend(self):
        return self._backend


from ast import Attribute
from .validators import TypeOf, OneOf, Bool


class PackageConfig:
    __slots__ = (
        "_debug",
        "_semaphore",
        "_raise_errors",
        "_return_callbacks",
        "_run_callbacks_on_error",
        "_print_callback_exceptions",
        "_print_callback_tracebacks",
        "_threaded_timeout",
    )

    debug:                          Bool                    = Bool()
    semaphore:                      OneOf                   = OneOf("local", "global")
    raise_errors:                   Bool                    = Bool()
    return_callbacks:               Bool                    = Bool()
    run_callbacks_on_error:         Bool                    = Bool()
    print_callback_exceptions:      Bool                    = Bool()
    print_callback_tracebacks:      Bool                    = Bool()
    threaded_timeout:               TypeOf                  = TypeOf(int, float, None.__class__)

    def __init__(
        self,
        debug:                      bool                    = False,
        semaphore:                  str                     = "local",
        raise_errors:               bool                    = True,
        return_callbacks:           bool                    = False,
        run_callbacks_on_error:     bool                    = False,
        print_callback_exceptions:  bool                    = True,
        print_callback_tracebacks:  bool                    = False,
        threaded_timeout:           int | float | None      = None,
    ):
        self.debug                                          = debug
        self.semaphore                                      = semaphore
        self.raise_errors                                   = raise_errors
        self.return_callbacks                               = return_callbacks
        self.run_callbacks_on_error                         = run_callbacks_on_error
        self.print_callback_exceptions                      = print_callback_exceptions
        self.print_callback_tracebacks                      = print_callback_tracebacks
        self.threaded_timeout                               = threaded_timeout

    def __repr__(self):
        keys = tuple(key.lstrip('_') for key in self.__slots__)
        sep = "="
        max_var_length = max(len(key) for key in keys)
        max_value_length = max(len(str(getattr(self, key))) for key in keys)
        max_line_length = max_var_length + max_value_length + 3
        line_sep = '-' * max_line_length
        return f"<Sessions Package Config>\n{line_sep}\n{"\n".join(f'{key} {sep:>{max_var_length - len(key) + 1}} {getattr(self, key)}' for key in keys)}\n{line_sep}"

    def __str__(self):
        return repr(self)

    def set(self, key, value):
        if f"_{key}" in self.__slots__:
            return setattr(self, key, value)
        raise AttributeError(f"SessionConfig has no option: '{key}'")

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        if f"_{key}" in self.__slots__:
            return setattr(self, key, value)
        raise AttributeError(f"SessionConfig has no option: '{key}'")

    def __getattr__(self, __name: str):
       print(f"SessionConfig has no option: '{__name}'")

SessionConfig = PackageConfig()


import os as _os
from random import Random as _Random
from pickle import (
    load as _load,
    dump as _dump,
    HIGHEST_PROTOCOL as _PICKLE_HIGHEST_PROTOCOL
)

from dotenv import dotenv_values as _dotenv_values
from requests import Session
from requests.auth import HTTPProxyAuth
from urllib.parse import (
    quote as _quote,
    unquote as _unquote,
)

from .useragents import UserAgents as _UserAgents

_MODULE_DIR = _os.path.dirname(_os.path.abspath(__file__))
_NORDVPN_SERVER_URL = "https://nordvpn.com/api/server"


def _path(relative):
    return _os.path.join(_MODULE_DIR, relative)

try:
    with open(_path("proxies.tuple"), "rb") as f:
        _PROXIES = _load(f)
except:
    with Session() as session:
        response = session.get(_NORDVPN_SERVER_URL)
    servers = response.json()
    _PROXIES = tuple(server["domain"] for server in servers if server["features"]["proxy_ssl"])
    with open(_path("proxies.tuple"), "wb") as f:
        _dump(_PROXIES, f, protocol=_PICKLE_HIGHEST_PROTOCOL)


class ProxySession(Session):
    def __init__(self, username="", password="", headers=None, random_user_agents=True, **kwargs):
        headers = headers if isinstance(headers, dict) else {}
        super().__init__()
        self._rng = _Random()
        self._random_user_agents = random_user_agents
        env = _dotenv_values()
        self._username = _quote(_unquote(username)) or env.get("PROXY_USERNAME", env.get("NORDVPN_USERNAME", ""))
        self._password = _quote(_unquote(password)) or env.get("PROXY_PASSWORD", env.get("NORDVPN_PASSWORD", ""))
        self._auth = HTTPProxyAuth(self._username, self._password)


    @classmethod
    def update_proxies(cls):
        global _PROXIES
        with Session() as session:
            response = session.get(_NORDVPN_SERVER_URL)
        servers = response.json()
        _PROXIES = tuple(server["domain"] for server in servers if server["features"]["proxy_ssl"])
        with open(_path("proxies.tuple"), "wb") as f:
            _dump(_PROXIES, f, protocol=_PICKLE_HIGHEST_PROTOCOL)


    @property
    def proxies(self):
        #url = f"{self._username}:{self._password}@{self._rng.choice(_PROXIES)}"
        url = self._rng.choice(_PROXIES)
        return {
            "http": f"http://{url}:89",
            "https": f"https://{url}:89"
        }


    @proxies.setter
    def proxies(self, value):
        pass


    def request(self, method, url, headers=None, **kwargs):
        proxies = self.proxies
        print(proxies)
        if self._random_user_agents:
            headers = headers or {}
            headers["User-Agent"] = _UserAgents.user_agent
        try:
            return super().request(method=method, url=url, proxies=proxies, headers=headers, auth=self._auth, **kwargs)
        except Exception as e:
            print(e)
            return self.request(method=method, url=url, **kwargs)


    def get(self, url, **kwargs):
        return self.request(method="GET", url=url, **kwargs)


    def head(self, url, **kwargs):
        return self.request(method="HEAD", url=url, **kwargs)


    def options(self, url, **kwargs):
        return self.request(method="OPTIONS", url=url, **kwargs)


    def delete(self, url, **kwargs):
        return self.request(method="DELETE", url=url, **kwargs)


    def post(self, url, **kwargs):
        return self.request(method="POST", url=url, **kwargs)


    def put(self, url, **kwargs):
        return self.request(method="PUT", url=url, **kwargs)


    def patch(self, url, **kwargs):
        return self.request(method="PATCH", url=url, **kwargs)


from dataclasses import make_dataclass
from time import time

from .abstract import Ratelimit, RatelimitDecoratorMixin


class TokenBucket(Ratelimit):
    """
    A token bucket implementation for rate limiting.

    Args:
        capacity (float | int): The maximum number of tokens the bucket can hold.
        fill_rate (float | int): The rate at which tokens are added to the bucket per second.
        **kwargs: Additional keyword arguments to be passed to the parent class.

    Attributes:
        _capacity (float | int): The maximum number of tokens the bucket can hold.
        _fill_rate (float | int): The rate at which tokens are added to the bucket per second.
    """

    __slots__ = ("_capacity", "_fill_rate")
    __dc__ = make_dataclass("TokenBucketCache", (("tokens", float), ("last_check", float), ("last_update", float)), slots=True, eq=False)

    def __init__(
        self,
        capacity: float | int           = 10,
        fill_rate: float | int          = 1,
        key: str | None                 = None,
        **kwargs
    ):
        super().__init__(backend="memory", key=key, **kwargs)
        self._capacity = capacity
        self._fill_rate = fill_rate


    def default(self):
        """
        Returns the default token bucket cache.

        Returns:
            TokenBucketCache: The default token bucket cache with tokens set to the capacity and last_check and last_update set to the current time.
        """
        now = self.now
        return self.__dc__(tokens=self._capacity, last_check=now, last_update=now)


    def get_tokens(self, key):
        """
        Retrieves the token bucket cache for the given key and updates the number of tokens based on the elapsed time and fill rate.

        Args:
            key: The key associated with the token bucket cache.

        Returns:
            TokenBucketCache: The updated token bucket cache.
        """
        # Retrieve data
        data = self._instance._ratelimit_memory_conn[key]

        # Calculate the time elapsed since the last fill
        current = time()
        elapsed = current - data.last_check # type: ignore

        # Calculate the number of tokens to add based on the fill rate
        to_add = elapsed * (self._fill_rate / self._capacity)

        # Set the new number of tokens (up to capacity)
        data.tokens = min(self._capacity, data.tokens + to_add) # type: ignore
        data.last_check = current # type: ignore
        return data


    def ok(self, key):
        """
        Checks if there are enough tokens in the bucket for the given key and updates the token bucket cache accordingly.

        Args:
            key: The key associated with the token bucket cache.

        Returns:
            bool: True if there are enough tokens and False otherwise.
        """
        data = self.get_tokens(key)
        if data.tokens >= 1: # type: ignore
            data.tokens -= 1 # type: ignore
            data.last_update = time() # type: ignore
            return True
        data.last_update = time() # type: ignore
        return False


# The `LeakyBucket` class is a decorator that implements a rate limit for a given function
# using the leaky bucket algorithm.
class LeakyBucket(Ratelimit):
    """
    LeakyBucket class represents a rate limiter based on the leaky bucket algorithm.

    Args:
        capacity (float | int): The maximum number of requests allowed within a certain time period.
        leak_rate (float | int): The rate at which the bucket leaks requests per second.
        **kwargs: Additional keyword arguments to be passed to the parent class.

    Attributes:
        _capacity (float | int): The maximum number of requests allowed within a certain time period.
        _leak_rate (float | int): The rate at which the bucket leaks requests per second.
    """

    __slots__ = ("_capacity", "_leak_rate")
    __dc__ = make_dataclass("LeakyBucketCache", (("content", float), ("last_checked", float), ("last_update", float)), slots=True, eq=False)

    def __init__(
        self,
        capacity: float | int           = 10,
        leak_rate: float | int          = 5,
        key: str | None                 = None,
        **kwargs
    ):
        super().__init__(backend="memory", key=key, **kwargs)
        self._capacity = capacity
        self._leak_rate = leak_rate


    def default(self):
        """
        Returns the default cache data for a key.

        Returns:
            LeakyBucketCache: The default cache data for a key.
        """
        now = self.now
        return self.__dc__(content=0, last_checked=now, last_update=now)


    def _leak(self, key):
        """
        Leaks requests from the bucket based on the elapsed time.

        Args:
            key: The key associated with the cache data.

        Returns:
            LeakyBucketCache: The updated cache data after leaking requests.
        """
        data = self._instance._ratelimit_memory_conn[key]
        content = data.content # type: ignore

        # Calculate the amount of time that has passed
        current_time = time()
        elapsed = current_time - data.last_checked # type: ignore
        data.last_checked = current_time # type: ignore

        # Leak the appropriate amount of requests
        data.content -= elapsed * self._leak_rate # type: ignore
        data.content = max(data.content, 0) # type: ignore
        return data


    def ok(self, key):
        """
        Checks if a request is allowed based on the leaky bucket algorithm.

        Args:
            key: The key associated with the cache data.

        Returns:
            bool: True if the request is allowed, False otherwise.
        """
        data = self._leak(key)
        if data.content < self._capacity: # type: ignore
            data.content += 1 # type: ignore
            data.last_update = time() # type: ignore
            return True
        data.last_update = time() # type: ignore
        return False


class SlidingWindow(Ratelimit):
    __slots__ = ("_limit", "_window")
    __dc__ = make_dataclass("SlidingWindowCache", (("pre_count", float), ("cur_count", float), ("cur_time", float), ("last_update", float)), slots=True, eq=False)

    def __init__(
        self,
        limit: int                      = 10,
        window: float | int             = 1.0,
        key: str | None                 = None,
        **kwargs
    ):
        super().__init__(backend="memory", key=key, **kwargs)
        self._limit = limit
        self._window = window


    def default(self):
        now = self.now
        return self.__dc__(pre_count=self._limit, cur_count=0, cur_time=now, last_update=now)


    def ok(self, key):
        data = self._instance._ratelimit_memory_conn[key]
        if ((time_ := time()) - data.cur_time) > self._window: # type: ignore
            data.cur_time = time_ # type: ignore
            data.pre_count = data.cur_count # type: ignore
            data.cur_count = 0 # type: ignore

        ec = (data.pre_count * (self._window - (time() - data.cur_time)) / self._window) + data.cur_count # type: ignore
        if ec < self._limit:
            data.cur_count += 1 # type: ignore
            data.last_update = time() # type: ignore
            return True
        data.last_update = time() # type: ignore
        return False


class FixedWindow(Ratelimit):
    __slots__ = ("_limit", "_window")
    __dc__ = make_dataclass("FixedWindowCache", (("window_start", float), ("requests", float), ("last_update", float)), slots=True, eq=False)

    def __init__(
        self,
        limit: int                      = 10,
        window: float | int             = 10.0,
        key: str | None                 = None,
        **kwargs
    ):
        super().__init__(backend="memory", key=key, **kwargs)
        self._window = window
        self._limit = limit


    def default(self):
        now = self.now
        return self.__dc__(window_start=now, requests=0, last_update=now)


    def ok(self, key):
        data = self._instance._ratelimit_memory_conn[key]
        window_start = data.window_start # type: ignore
        current_time = time()
        if current_time - window_start > self._window:
            data.requests = 0 # type: ignore
            data.window_start = current_time # type: ignore

        if data.requests < self._limit: # type: ignore
            data.requests += 1 # type: ignore
            data.last_update = time() # type: ignore
            return True
        data.last_update = time() # type: ignore
        return False


# The `GCRA` class is a decorator that limits the rate at which a function can be called
# based on a specified rate and burst size.
class GCRA(Ratelimit):
    """
    GCRA (Generic Cell Rate Algorithm) is a rate limiting algorithm that allows a burst of requests
    up to a certain limit within a specified time period.

    Args:
        period (float | int): Time period for each cell/token (in seconds). Default is 1.0.
        limit (int): Limit on the burst size (in seconds). Default is 10.
        **kwargs: Additional keyword arguments to be passed to the parent class.

    Attributes:
        _period (float | int): Time period for each cell/token (in seconds).
        _limit (int): Limit on the burst size (in seconds).
    """

    __slots__ = ("_period", "_limit")
    __dc__ = make_dataclass("GCRACache", (("last_time", float), ("last_update", float)), slots=True, eq=False)

    def __init__(
        self,
        period: float | int     = 1.0,
        limit: int              = 10,
        key: str | None         = None,
        **kwargs
    ):
        super().__init__(backend="memory", key=key, **kwargs)
        self._period = period  # Time period for each cell/token (in seconds)
        self._limit = limit  # Limit on the burst size (in seconds)


    def default(self):
        """
        Returns the default cache data object.

        Returns:
            GCRACache: The default cache data object with last_time and last_update set to the current time.
        """
        now = self.now
        return self.__dc__(last_time=now, last_update=now)


    def ok(self, key):
        """
        Checks if a request with the given key is allowed based on the rate limit.

        Args:
            key: The key associated with the request.

        Returns:
            bool: True if the request is allowed, False otherwise.
        """
        data = self._instance._ratelimit_memory_conn[key]
        current_time = time()
        expected_time = data.last_time + self._period

        if current_time < expected_time - self._limit:
            # The cell/token arrives too early and does not conform.
            data.last_update = current_time # type: ignore
            return False
        else:
            # The cell/token conforms; update the last_time.
            data.last_time = max(expected_time, current_time) # type: ignore
            data.last_update = current_time # type: ignore
            return True


_CLASS_TYPES = {
    "leakybucket": LeakyBucket,
    "tokenbucket": TokenBucket,
    "slidingwindow": SlidingWindow,
    "fixedwindow": FixedWindow,
    "gcra": GCRA,
}

class PyRatelimitDecoratorFactory:
    _type = type

    def __new__(cls, type="slidingwindow", *args, **kwargs):
        type_name = type
        new_type = cls._type(type, (_CLASS_TYPES[type_name], RatelimitDecoratorMixin), {})
        self = new_type.__new__(new_type)
        self.__init__(*args, **kwargs)
        return self


class PyRatelimitFactory:
    def __new__(cls, instance, type="slidingwindow", *args, **kwargs):
        type_name = type
        type = _CLASS_TYPES[type_name]
        self = type.__new__(type)
        self._instance = instance
        self.__init__(*args, **kwargs)
        self._ratelimit_type = type_name
        return self


import sqlite3
from time import time
from dataclasses import make_dataclass, field, astuple
from functools import partial

from .abstract import Ratelimit


class TokenBucket(Ratelimit):
    """
    A token bucket implementation for rate limiting.

    Args:
        capacity (float | int): The maximum number of tokens the bucket can hold. Defaults to 10.
        fill_rate (float | int): The rate at which tokens are added to the bucket per second. Defaults to 1.
        **kwargs: Additional keyword arguments to be passed to the parent class.

    Attributes:
        _capacity (float | int): The maximum number of tokens the bucket can hold.
        _fill_rate (float | int): The rate at which tokens are added to the bucket per second.
    """

    __slots__ = ("_capacity", "_fill_rate")
    __attrs__ = ("key", "tokens", "last_check", "expiration")
    __dc__ = make_dataclass("TokenBucketCache", fields=(("key", str), ("tokens", float), ("last_check", float), ("expiration", float)), slots=True, eq=False)

    def __init__(
        self,
        capacity: float | int           = 10,
        fill_rate: float | int          = 1,
        key: str | None                 = None,
        conn: sqlite3.Connection | None = None,
        **kwargs
    ):
        super().__init__(backend="sqlite", key=key, conn=conn, **kwargs)
        self._capacity = capacity
        self._fill_rate = fill_rate

    def default(self):
        """
        Returns the default token bucket data.

        Returns:
            dict: The default token bucket data, including the initial number of tokens, last check time, and expiration time.
        """
        return dict(
            tokens=self._capacity,
            last_check=self.now,
            expiration=time() + self.options.cache_timeout if self.options.cache_timeout else 0
        )

    def get_tokens(self, key):
        """
        Retrieves the number of tokens from the token bucket for the given key.

        Args:
            key: The key associated with the token bucket.

        Returns:
            TokenBucketCache: The token bucket data, including the number of tokens, last check time, and expiration time.
        """
        # Retrieve data
        data = self[key]

        # Calculate the time elapsed since the last fill
        current = time()
        elapsed = current - data.last_check # type: ignore

        # Calculate the number of tokens to add based on the fill rate
        to_add = elapsed * (self._fill_rate / self._capacity)

        # Set the new number of tokens (up to capacity)
        data.tokens = min(self._capacity, data.tokens + to_add) # type: ignore
        data.last_check = current # type: ignore
        return data


    def ok(self, key):
        """
        Checks if a token can be consumed from the token bucket for the given key.

        Args:
            key: The key associated with the token bucket.

        Returns:
            bool: True if a token can be consumed, False otherwise.
        """
        output = False
        data = self.get_tokens(key)

        if data.tokens >= 1: # type: ignore
            data.tokens -= 1 # type: ignore
            output = True

        self[key] = data
        return output


# The `LeakyBucket` class is a decorator that implements a rate limit for a given function
# using the leaky bucket algorithm.
class LeakyBucket(Ratelimit):
    """
    A class representing a leaky bucket rate limiter.

    The leaky bucket algorithm is a rate limiting algorithm that allows a certain number of requests
    to be processed per unit of time. If the bucket is full, additional requests are either discarded
    or delayed.

    Args:
        capacity (float | int): The maximum capacity of the bucket.
        leak_rate (float | int): The rate at which the bucket leaks requests.

    Attributes:
        _capacity (float | int): The maximum capacity of the bucket.
        _leak_rate (float | int): The rate at which the bucket leaks requests.
    """

    __slots__ = ("_capacity", "_leak_rate")
    __attrs__ = ("key", "content", "last_check", "expiration")
    __dc__ = make_dataclass("LeakyBucketCache", fields=(("key", str), ("content", float), ("last_check", float), ("expiration", float, field(default=0.0))), slots=True, eq=False)

    def __init__(
        self,
        capacity: float | int           = 10,
        leak_rate: float | int          = 5,
        key: str | None                 = None,
        conn: sqlite3.Connection | None = None,
        **kwargs
    ):
        super().__init__(backend="sqlite", key=key, conn=conn, **kwargs)
        self._capacity = capacity
        self._leak_rate = leak_rate

    def default(self):
        return dict(
            content=0,
            last_check=self.now,
            expiration=time() + self.options.cache_timeout if self.options.cache_timeout else 0
        )

    def _leak(self, key):
        """
        Leaks the appropriate amount of requests from the specified key.

        Args:
            key: The key to leak requests from.

        Returns:
            The updated data object after leaking requests.
        """
        data = self[key]

        # Calculate the amount of time that has passed
        current_time = time()
        elapsed = current_time - data.last_check # type: ignore
        data.last_check = current_time # type: ignore

        # Leak the appropriate amount of requests
        data.content -= elapsed * self._leak_rate # type: ignore
        data.content = max(data.content, 0) # type: ignore
        return data


    def ok(self, key):
        """
        Checks if the given key is within the rate limit capacity and updates the count if it is.

        Args:
            key (str): The key to check against the rate limit.

        Returns:
            bool: True if the key is within the rate limit capacity and the count is updated, False otherwise.
        """
        output = False
        data = self._leak(key)

        if data.content < self._capacity: # type: ignore
            data.content += 1 # type: ignore
            output = True

        self[key] = data
        return output


class SlidingWindow(Ratelimit):
    """
    A sliding window rate limiter implementation.

    This class provides rate limiting functionality based on a sliding window algorithm.
    It tracks the number of requests made within a specified time window and enforces a limit on the number of requests allowed.

    Attributes:
        _limit (int): The maximum number of requests allowed within the time window.
        _window (float | int): The time window (in seconds) within which the requests are counted.

    Methods:
        __init__(self, limit: int = 10, window: float | int = 1.0, **kwargs): Initializes the SlidingWindow rate limiter.
        default(self): Returns the default rate limit data for a key.
        ok(self, key): Checks if the given key is allowed to proceed based on the rate limit rules.

    Example usage:
        sliding_window = SlidingWindow(limit=10, window=1.0)
        if sliding_window.ok("user123"):
            # Allow the request to proceed
        else:
            # Rate limit exceeded, reject the request
    """

    __slots__ = ("_limit", "_window")
    __attrs__ = ("key", "cur_time", "cur_count", "pre_count", "expiration")
    __dc__ = make_dataclass("SlidingWindowCache", fields=(("key", str), ("cur_time", float), ("cur_count", int), ("pre_count", int), ("expiration", float, field(default=0.0))), slots=True, eq=False)

    def __init__(
        self,
        limit: int                      = 10,
        window: float | int             = 1.0,
        key: str | None                 = None,
        conn: sqlite3.Connection | None = None,
        **kwargs
    ):
        super().__init__(backend="sqlite", key=key, conn=conn, **kwargs)
        self._limit = limit
        self._window = window

    def default(self):
        return dict(
            cur_time=self.now,
            cur_count=0,
            pre_count=self._limit,
            expiration=time() + self.options.cache_timeout if self.options.cache_timeout else 0
        )

    def ok(self, key):
        """
        Checks if the given key is allowed to proceed based on the rate limit rules.

        Args:
            key: The key to check for rate limiting.

        Returns:
            bool: True if the key is allowed to proceed, False otherwise.
        """
        output = False
        data = self[key]

        if ((time_ := time()) - data.cur_time) > self._window: # type: ignore
            data.cur_time = time_ # type: ignore
            data.pre_count = data.cur_count # type: ignore
            data.cur_count = 0 # type: ignore

        ec = (data.pre_count * (self._window - (time() - data.cur_time)) / self._window) + data.cur_count # type: ignore

        if ec < self._limit:
            data.cur_count += 1 # type: ignore
            output = True

        self[key] = data
        return output


class FixedWindow(Ratelimit):
    """
    Represents a fixed window rate limiter.

    Attributes:
        _limit (int): The maximum number of requests allowed within the window.
        _window (float | int): The duration of the window in seconds.
    """

    __slots__ = ("_limit", "_window")
    __attrs__ = ("key", "window_start", "requests", "expiration")
    __dc__ = make_dataclass("FixedWindowCache", fields=(("key", str), ("window_start", float), ("requests", int), ("expiration", float, field(default=0.0))), slots=True, eq=False)

    def __init__(
        self,
        limit: int                      = 10,
        window: float | int             = 10.0,
        key: str | None                 = None,
        conn: sqlite3.Connection | None = None,
        **kwargs
    ):
        super().__init__(backend="sqlite", key=key, conn=conn, **kwargs)
        self._window = window
        self._limit = limit

    def default(self):
        return dict(
            window_start=self.now,
            requests=0,
            expiration=time() + self.options.cache_timeout if self.options.cache_timeout else 0
        )

    def ok(self, key):
        """
        Checks if the given key is allowed to make a request based on the rate limit rules.

        Args:
            key: The key to check for rate limiting.

        Returns:
            True if the key is allowed to make a request, False otherwise.
        """
        output = False
        data = self[key]
        window_start = data.window_start # type: ignore
        current_time = time()

        if current_time - window_start > self._window:
            data.requests = 0 # type: ignore
            data.window_start = current_time # type: ignore

        if data.requests < self._limit: # type: ignore
            data.requests += 1 # type: ignore
            output = True

        self[key] = data
        return output


# The `GCRA` class is a decorator that limits the rate at which a function can be called
# based on a specified rate and burst size.
class GCRA(Ratelimit):
    """
    GCRA (Generic Cell Rate Algorithm) is a rate limiting algorithm that limits the rate at which requests can be made.
    It is implemented using a SQLite backend.

    Attributes:
        _period (float | int): Time period for each cell/token (in seconds).
        _limit (int): Limit on the burst size (in seconds).

    Args:
        period (float | int, optional): Time period for each cell/token (in seconds). Defaults to 1.0.
        limit (int, optional): Limit on the burst size (in seconds). Defaults to 10.
        **kwargs: Additional keyword arguments to be passed to the parent class.

    Methods:
        default(): Returns the default data structure for a new key.
        ok(key): Checks if a request with the given key is allowed based on the rate limit.

    """

    __slots__ = ("_period", "_limit")
    __attrs__ = ("key", "last_check", "expiration")
    __dc__ = make_dataclass("GCRACache", fields=(("key", str), ("last_check", float), ("expiration", float, field(default=0.0))), slots=True, eq=False)

    def __init__(
        self,
        period: float | int             = 1.0,
        limit: int                      = 10,
        key: str | None                 = None,
        conn: sqlite3.Connection | None = None,
        **kwargs
    ):
        super().__init__(backend="sqlite", key=key, conn=conn, **kwargs)
        self._period = period  # Time period for each cell/token (in seconds)
        self._limit = limit  # Limit on the burst size (in seconds)

    def default(self):
        """
        Returns the default data structure for a new key.

        Returns:
            dict: The default data structure containing the last_check and expiration values.
        """
        return dict(
            last_check=self.now,
            expiration=time() + self.options.cache_timeout if self.options.cache_timeout else 0
        )

    def ok(self, key):
        """
        Checks if a request with the given key is allowed based on the rate limit.

        Args:
            key (str): The key associated with the request.

        Returns:
            bool: True if the request is allowed, False otherwise.
        """
        output = False
        data = self[key]
        current_time = time()
        expected_time = data.last_check + self._period

        if not current_time < expected_time - self._limit:
            # The cell/token conforms; update the last_check.
            data.last_check = max(expected_time, current_time) # type: ignore
            output = True

        self[key] = data
        return output


_CLASS_TYPES = {
    "leakybucket": LeakyBucket,
    "tokenbucket": TokenBucket,
    "slidingwindow": SlidingWindow,
    "fixedwindow": FixedWindow,
    "gcra": GCRA,
}



class SQLiteRatelimitFactory:
    """
    Factory class for creating SQLite ratelimit objects.

    This class provides methods for interacting with a SQLite database to implement ratelimiting functionality.
    """
    _type = type

    def __new__(cls, instance, type="slidingwindow", key=None, *args, **kwargs):
        type_name = type
        type = _CLASS_TYPES[type_name]
        type = cls._type(type.__name__, (type, ), {
            "__slots__": ("_cursor", "_getstatement", "_setstatement", "_create_tables", "__contains__", "__getitem__", "__setitem__", "__delitem__", "clear", "keys", "values", "items", "_cleanup"),
        })

        self = type.__new__(type)
        self._getstatement = f"SELECT {', '.join(attr for attr in self.__attrs__)} FROM ratelimit WHERE key = ?"
        self._setstatement = f"INSERT OR REPLACE INTO ratelimit ({', '.join(attr for attr in self.__attrs__)}) VALUES (?{', ?'.join("" for _ in range(len(self.__attrs__)))})"
        self._instance = instance
        self._ratelimit_type = type_name
        self._create_tables = partial(SQLiteRatelimitFactory._create_tables, self)
        self.__contains__ = partial(SQLiteRatelimitFactory.__contains__, self)
        self.__getitem__ = partial(SQLiteRatelimitFactory.__getitem__, self)
        self.__setitem__ = partial(SQLiteRatelimitFactory.__setitem__, self)
        self.__delitem__ = partial(SQLiteRatelimitFactory.__delitem__, self)
        self.clear = partial(SQLiteRatelimitFactory.clear, self)
        self.keys = partial(SQLiteRatelimitFactory.keys, self)
        self.values = partial(SQLiteRatelimitFactory.values, self)
        self.items = partial(SQLiteRatelimitFactory.items, self)
        self._cleanup = partial(SQLiteRatelimitFactory._cleanup, self)
        self.__init__(*args, key=key, **kwargs)
        return self


    def _create_tables(self):
        with self._instance._sqlite_conn:
            self._instance._sqlite_conn.execute("""
                CREATE TABLE IF NOT EXISTS ratelimit (
                key TEXT PRIMARY KEY,
                requests INT,
                window_start FLOAT,
                cur_time FLOAT,
                cur_count INT,
                pre_count INT,
                tokens FLOAT,
                last_check FLOAT,
                content FLOAT,
                expiration FLOAT
                )
            """)

    def __contains__(self, url):
        key = self._parse_key(url)
        self._cursor.execute("SELECT key FROM ratelimit WHERE key = ?", (key,))
        return bool(self._cursor.fetchone())

    def __getitem__(self, key):
        self._cursor.execute(self._getstatement, (key,))
        data = self._cursor.fetchone()
        if data is not None:
            if self.options.cache_timeout and data[-1] < time():
                del self[key]
                return self.__dc__(key=key, **self.default())
            return self.__dc__(*data)
        return self.__dc__(key=key, **self.default())


    def __setitem__(self, key, value):
        value.expiration = time() + self.options.cache_timeout if self.options.cache_timeout else 0
        with self._instance._sqlite_conn:
            self._instance._sqlite_conn.execute(self._setstatement, astuple(value))

    def __delitem__(self, key):
        with self._instance._sqlite_conn:
            self._instance._sqlite_conn.execute("DELETE FROM ratelimit WHERE key = ?", (key,))

    def clear(self):
        with self._instance._sqlite_conn:
            self._instance._sqlite_conn.execute("DELETE FROM ratelimit")

    def keys(self):
        self._cursor.execute("SELECT key FROM ratelimit")
        return tuple(key[0] for key in self._cursor.fetchall())

    def values(self):
        self._cursor.execute("SELECT value FROM ratelimit")
        return tuple(value[0] for value in self._cursor.fetchall())

    def items(self):
        self._cursor.execute("SELECT key, value FROM ratelimit")
        return tuple((key, value) for key, value in self._cursor.fetchall())

    def _cleanup(self):
        pass


from .abstract import Cache

class RedisCache(Cache):
    __slots__ = ()

    def __init__(
        self,
        instance: object,
        **kwargs
    ):
        self._instance = instance
        super().__init__(backend="redis", **kwargs)

    def __contains__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)
        return self._instance._redis_conn.exists(key)

    @Cache.deserialize
    def __getitem__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)

        value = self._instance._redis_conn.getex(key, int(self.options.cache_timeout))
        if value is None:
            return value

        if self.options.compression:
            return self._decompress(value)

        return value

    @Cache.serialize
    def __setitem__(self, key, value):
        if not key.endswith(":cache"):
            key = self._parse_key(key)
        if self.options.compression:
            value = self._compress(value)

        self._instance._redis_conn.setex(key, int(self.options.cache_timeout), value)

    def __delitem__(self, key):
        if not key.endswith(":cache"):
            key = self._parse_key(key)
        return self._instance._redis_conn.delete(key)

    def clear(self):
        return self._instance._redis_conn.flushdb()

    def keys(self):
        if self.options.decode_responses:
            return tuple(key for key in self._instance._redis_conn.keys() if key.endswith("cache"))
        else:
            return tuple(key for key in self._instance._redis_conn.keys() if key.endswith(b"cache"))

    def values(self):
        if self.options.decode_responses:
            return tuple(self._instance._redis_conn.get(key) for key in self.keys() if key.endswith("cache"))
        else:
            return tuple(self._instance._redis_conn.get(key) for key in self.keys() if key.endswith(b"cache"))

    def items(self):
        return tuple(zip(self.keys(), self.values()))

    def _cleanup(self):
        pass


from setuptools import setup, find_packages

setup(
    name="sessions",
    version="0.1.1",
    packages=find_packages(),
    install_requires=[
        "orjson",
        "requests",
        "httpx[http2]>=0.14.0",
        "aiohttp[speedups]>=3.9",
        "alive_progress",
        "python-dotenv",
        "yarl",
        "redislite",
        "redis",
    ],
    extras_require={
        "all": ["redislite", "stem"],
        "backend": ["redislite"],
        "proxy": ["requests[socks]"],
        "tor": ["stem"],
    }
)